#if !defined BSDF_TEMPLATE
#define BSDF_TEMPLATE
    /*
        Calculate the Lambertian diffuse BSDF.

        Just a generic Lambertian diffuse BSDF, not much else to say. This doesn't actually need phase functions or facet BSDFs, it has them for 
        the sake of consistency.
    */

    vec3 SampleLambertianPhaseFunction(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        throughput *= bounceMaterial.albedo * bounceMaterial.ao;
        estimate += throughput * bounceMaterial.emission;
        return GenerateCosineVector(vec3(0, 0, 1), RandNext2F());
    }

    vec3 SampleLambertian(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        return SampleLambertianPhaseFunction(direction, estimate, throughput, wavelength, bounceMaterial);
    }

    float EstimateLambertianPhaseFunction(in vec3 incomingDirection, in vec3 outgoingDirection, in Material bounceMaterial) {
        return (bounceMaterial.albedo / pi) * saturate(outgoingDirection.z);
    }

    float EstimateLambertian(in vec3 incomingDirection, in vec3 outgoingDirection, in Material bounceMaterial) {
        return EstimateLambertianPhaseFunction(incomingDirection, outgoingDirection, bounceMaterial);
    }
#endif