#if !defined BSDF_NORMAL_MAPPED
#define BSDF_NORMAL_MAPPED
    /*
       Calculates the normal mapped and microfacet based normal mapping BSDF for all nodes.
    */

    vec3 SampleOpaqueDielectric_NM(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        return tbnMatrixP * SampleOpaqueSelector(direction * tbnMatrixP, estimate, throughput, wavelength, bounceMaterial);
    }

    vec3 SampleOpaqueDielectric_MBN(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        bounceMaterial.clearCoat = 0.0;

        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        vec3 wm;
        bool intersectWP;
        if(RandNextF() < lambda_p(wp, -direction)) {
            wm = wp;
            intersectWP = true;
        } else {
            wm = wt(wp);
            intersectWP = false;
        }

        mat3 tbnMatrixT = GetRotationMatrix(vec3(0, 0, 1), wt(wp));

        int i = 0;
        while(i <= MAX_MICROFACET_BASED_NORMAL_MAP_PATH_SEGMENTS) {
            i += 1;

            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            vec3 newDirection = vec3(0.0);
            newDirection = (intersectWP ? tbnMatrixP : tbnMatrixT) * SampleOpaqueSelector(direction * (intersectWP ? tbnMatrixP : tbnMatrixT), estimate, throughput, wavelength, bounceMaterial);

            direction = newDirection;

            if(throughput == 0.0) {
                return direction;
            }

            if(RandNextF() < G1_MN(direction, wp, wm)) {
                break;
            } else {
                wm = intersectWP ? wt(wp) : wp;
                intersectWP = !intersectWP;
            }                
        }

        return direction;
    }

    float EstimateOpaqueDielectric_NM(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounceID, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        return EstimateOpaqueDielectric(incomingDirection * tbnMatrixP, outgoingDirection * tbnMatrixP, bounceID, wavelength, bounceMaterial);
    }

    float EstimateOpaqueDielectric_MBN(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounceID, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        bounceMaterial.clearCoat = 0.0;
        if(bounceMaterial.emissive) {
            bounceMaterial.clearCoat = 0.0;
        }
        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        vec3 wm;
        bool intersectWP;
        if(RandNextF() < lambda_p(wp, -incomingDirection)) {
            wm = wp;
            intersectWP = true;
        } else {
            wm = wt(wp);
            intersectWP = false;
        }

        mat3 tbnMatrixT = GetRotationMatrix(vec3(0, 0, 1), wt(wp));

        float estimate = 0.0;
        float throughput = 1.0;

        int i = 0;
        while(i <= MAX_MICROFACET_BASED_NORMAL_MAP_PATH_SEGMENTS) {
            i += 1;

            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            estimate += throughput * EstimateOpaqueDielectric(incomingDirection * (intersectWP ? tbnMatrixP : tbnMatrixT), outgoingDirection * (intersectWP ? tbnMatrixP : tbnMatrixT), bounceID, wavelength, bounceMaterial) * dot(outgoingDirection, wm) * G1_MN(outgoingDirection, wp, wm);

            float tmp;
            vec3 newDirection = vec3(0.0);
            newDirection = (intersectWP ? tbnMatrixP : tbnMatrixT) * SampleOpaqueDielectric(incomingDirection * (intersectWP ? tbnMatrixP : tbnMatrixT), tmp, throughput, wavelength, bounceMaterial);

            incomingDirection = newDirection;

            if(throughput == 0.0) {
                return estimate;
            }

            if(RandNextF() < G1_MN(incomingDirection, wp, wm)) {
                break;
            } else {
                wm = intersectWP ? wt(wp) : wp;
                intersectWP = !intersectWP;
            }                
        }

        return estimate;
    }

    vec3 SampleDielectric_NM(in vec3 direction, inout float throughput, in float wavelength, inout bool outside, in VoxelData data, in Material bounceMaterial) {
        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        return tbnMatrixP * SampleDielectric(direction * tbnMatrixP, throughput, outside, wavelength, bounceMaterial);
    }

    float EstimateDielectric_NM(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounceID, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        mat3 tbnMatrix2 = GetRotationMatrix(data.tbnMatrix[2], data.normalMap) * data.tbnMatrix;
        vec3 viewDirection   = incomingDirection * tbnMatrix2;
            viewDirection.z = sign(dot(incomingDirection, data.tbnMatrix[2])) * abs(viewDirection.z);

        return EstimateDielectric(viewDirection, outgoingDirection * tbnMatrix2, bounceID, wavelength, bounceMaterial) * (outgoingDirection * tbnMatrix2).z;
    }

    vec3 SampleConductor_NM(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        return tbnMatrixP * SampleConductor(direction * tbnMatrixP, throughput, wavelength, bounceMaterial);
    }

    vec3 SampleConductor_MBN(in vec3 direction, inout float throughput, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        bounceMaterial.clearCoat = 0.0;

        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        vec3 wm;
        bool intersectWP;
        if(RandNextF() < lambda_p(wp, -direction)) {
            wm = wp;
            intersectWP = true;
        } else {
            wm = wt(wp);
            intersectWP = false;
        }

        mat3 tbnMatrixT = GetRotationMatrix(vec3(0, 0, 1), wt(wp));

        int i = 0;
        while(i <= MAX_MICROFACET_BASED_NORMAL_MAP_PATH_SEGMENTS) {
            i += 1;

            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            vec3 newDirection = vec3(0.0);
            newDirection = (intersectWP ? tbnMatrixP : tbnMatrixT) * SampleConductor(direction * (intersectWP ? tbnMatrixP : tbnMatrixT), throughput, wavelength, bounceMaterial);

            direction = newDirection;

            if(throughput == 0.0) {
                return direction;
            }

            if(RandNextF() < G1_MN(direction, wp, wm)) {
                break;
            } else {
                wm = intersectWP ? wt(wp) : wp;
                intersectWP = !intersectWP;
            }                
        }

        return direction;
    }

    float EstimateConductor_NM(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounceID, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        return EstimateConductor(incomingDirection * tbnMatrixP, outgoingDirection * tbnMatrixP, bounceID, wavelength, bounceMaterial);
    }

    float EstimateConductor_MBN(in vec3 incomingDirection, in vec3 outgoingDirection, in int bounceID, in float wavelength, in VoxelData data, in Material bounceMaterial) {
        bounceMaterial.clearCoat = 0.0;

        vec3 wp = data.tangentNormal;
        mat3 tbnMatrixP = GetRotationMatrix(vec3(0, 0, 1), wp);

        vec3 wm;
        bool intersectWP;
        if(RandNextF() < lambda_p(wp, -incomingDirection)) {
            wm = wp;
            intersectWP = true;
        } else {
            wm = wt(wp);
            intersectWP = false;
        }

        mat3 tbnMatrixT = GetRotationMatrix(vec3(0, 0, 1), wt(wp));

        float estimate = 0.0;
        float throughput = 1.0;

        int i = 0;
        while(i <= MAX_MICROFACET_BASED_NORMAL_MAP_PATH_SEGMENTS) {
            i += 1;

            if(RandNextF() > throughput) {
                throughput = 0.0;
                break;
            }
            throughput /= saturate(throughput);

            estimate += throughput * EstimateConductor(incomingDirection * (intersectWP ? tbnMatrixP : tbnMatrixT), outgoingDirection * (intersectWP ? tbnMatrixP : tbnMatrixT), bounceID, wavelength, bounceMaterial) * dot(outgoingDirection, wm) * G1_MN(outgoingDirection, wp, wm);

            vec3 newDirection = vec3(0.0);
            newDirection = (intersectWP ? tbnMatrixP : tbnMatrixT) * SampleConductor(incomingDirection * (intersectWP ? tbnMatrixP : tbnMatrixT), throughput, wavelength, bounceMaterial);

            incomingDirection = newDirection;

            if(throughput == 0.0) {
                return estimate;
            }

            if(RandNextF() < G1_MN(incomingDirection, wp, wm)) {
                break;
            } else {
                wm = intersectWP ? wt(wp) : wp;
                intersectWP = !intersectWP;
            }                
        }

        return estimate;
    }
#endif