#if !defined BSDF_TEMPLATE
#define BSDF_TEMPLATE
    /*
        This file does not describe a BSDF, but it does give a simple overview of the BSDF format Chronos uses.

        A full BSDF should always include the estimate functions as well as the sampling functions. You can also include a microfacet based normal mapping
        estimate and sampling function for the BSDF, but it is not strictly required.
    */

    vec3 SampleTemplateFacetBSDF() {
        /*
            Calculate the facet BSDF to be used for sampling. All that would be included here is the throughput update, as well as the new ray direction.

            It can have any number of inputs, but the name should always follow the format of "SampleBSDFNameFacetBSDF".
        */
        return vec3(0.0);
    }

    vec3 SampleTemplate() {
        /*
            Sample the BSDF. This function is where the actual sampling happens, and it can be used to produce either a single scattering or 
            multiple scattering BSDF. This function is rarely ever just "return SampleBSDFNameFacetBSDF();", but it can be.

            It can have any number of inputs, but the name should always follow the format of "SampleBSDFName".
        */
        return vec3(0.0);
    }

    float EstimateTemplateFacetBSDF() {
        /*
            Calculate the facet BSDF to be used for estimation.

            It can have any number of inputs, but the name should always follow the format of "EstimateBSDFNameFacetBSDF".
        */
        return 0.0;
    }

    float EstimateTemplate() {
        /*
            Calculate the BSDF. This function is where the actual estimation of the BSDF happens, and it can be used to produce either a single scattering or 
            multiple scattering BSDF. This function is rarely ever just "return EstimateBSDFNameFacetBSDF();", but it can be. The "SampleBSDFNameFacetBSDF" 
            function will have to be used if doing multiple scattering.

            It can have any number of inputs, but the name should always follow the format of "EstimateBSDFName".
        */
        return 0.0;
    }
#endif