#if !defined BSDF_CHRONOS
#define BSDF_CHRONOS
    /*
        The Chronos Principled BSDF is designed with the purpose of being flexible and accurate without being insanely difficult to implement.
    */

    void DustMaterial(inout Material bounceMaterial, in float wavelength) {
        //I only need this for the time being
        float albedo = 0.0;

        vec3 dustColor = vec3(1.0, 0.0, 0.0) * rcp(pi4);

        RGBToSpectrum(albedo, wavelength, dustColor.r, dustColor.g, dustColor.b, 0);

        float k = mix(3e-3, 1.0e-3, (wavelength - 600.0) / 231.0);
        if(wavelength > 600.0) {
            k = 0.0;
        } else if(wavelength < 450.0) {
            k = mix(0.012, 3e-3, (wavelength - 450) / 381.0);
        }

        bounceMaterial.roughness = 0.0;
        bounceMaterial.albedo = albedo;
        bounceMaterial.n2 = complexFloat(mix(1.373, 1.341, (wavelength - 390.0) / 441.0), k);
    }

    void ChronosPrincipledSampled(
        in vec3 tangentRayDirection,
        in float wavelength,
        in VoxelData data,
        in Material bounceMaterial,

        inout vec3 outgoingRayDirection,
        inout float estimate,
        inout float throughput,
        inout Volume subsurface,
        inout bool isRefractRay
    ) {
        mat3 tbnMatrix2 = GetRotationMatrix(data.tbnMatrix[2], data.flatNormal) * data.tbnMatrix;

        if(!bounceMaterial.doRefract) {
            if(bounceMaterial.clearCoat == 0.0) {
                float conductorProbability = bounceMaterial.layerWeight;
                bool conductor = bounceMaterial.isMetal;
                if(conductor) {
                    outgoingRayDirection = tbnMatrix2 * SampleConductor_MBN(
                        tangentRayDirection,
                        throughput,
                        wavelength,
                        data,
                        bounceMaterial
                    );
                } else {
                    if(!bounceMaterial.isLambertian) {
                        outgoingRayDirection = tbnMatrix2 * SampleOpaqueDielectric_MBN(
                            tangentRayDirection,
                            estimate,
                            throughput,
                            wavelength,
                            data,
                            bounceMaterial
                        );
                    } else {
                        outgoingRayDirection = tbnMatrix2 * SampleLambertian(outgoingRayDirection, estimate, throughput, wavelength, bounceMaterial);
                    }
                }
            } else {
                outgoingRayDirection = tbnMatrix2 * SampleClearCoat(
                    tangentRayDirection,
                    estimate,
                    throughput,
                    wavelength,
                    data,
                    bounceMaterial
                );
            }

            throughput = saturate(throughput);

            if(isnan(throughput)) {
                throughput = 0.0;
            }
        } else {
            bool outside = true;

            outgoingRayDirection = tbnMatrix2 * SampleDielectric_NM(
                tangentRayDirection, 
                throughput, 
                wavelength, 
                outside, 
                data,
                bounceMaterial
            );

            isRefractRay = true;

            if(!outside) {
                subsurface = DefineSubsurface(bounceMaterial.albedo, bounceMaterial.radius, wavelength, data.id);
                subsurface.forwardG = bounceMaterial.subsurfaceForwardScatterG;
                subsurface.backwardG = bounceMaterial.subsurfaceBackwardScatterG;
            } else {
                subsurface = Empty;
                //if(data.sameMaterial) {
                //    subsurface = DefineSubsurface(bounceMaterial.albedo, bounceMaterial.radius, wavelength, data.id);
                //    subsurface.forwardG = bounceMaterial.subsurfaceForwardScatterG;
                //    subsurface.backwardG = bounceMaterial.subsurfaceBackwardScatterG;
                //}
            }

            if(isnan(throughput)) {
                throughput = 0.0;
            }
        }
    }

    void ChronosPrincipledEstimate(
        in vec3 incomingDirection,
        in vec3 outgoingDirection,
        in float wavelength,
        in VoxelData data,
        in Material bounceMaterial,
        in int bounce,

        inout float bsdf
    ) {
        mat3 tbnMatrix2 = GetRotationMatrix(data.tbnMatrix[2], data.flatNormal) * data.tbnMatrix;

        if(!bounceMaterial.doRefract) {
            float conductorProbability = bounceMaterial.layerWeight;
            bool conductor = bounceMaterial.isMetal;
            if(conductor) {
                bsdf += EstimateConductor_MBN(incomingDirection, outgoingDirection, bounce, wavelength, data, bounceMaterial);
            } else {
                if(!bounceMaterial.isLambertian) {
                    bsdf = EstimateOpaqueDielectric_MBN(incomingDirection, outgoingDirection, bounce, wavelength, data, bounceMaterial);
                } else {
                    bsdf = EstimateLambertian(incomingDirection, outgoingDirection, bounceMaterial);
                }
            }
        } else {
            bsdf = EstimateDielectric_NM(tbnMatrix2 * incomingDirection, tbnMatrix2 * outgoingDirection, 0, wavelength, data, bounceMaterial);
        }
    }
#endif