#if !defined SINGLE_SCATTER_BSDF_LIB
#define SINGLE_SCATTER_BSDF
    vec3 SampleGenericSingleScatteringBSDF(in vec3 direction, inout float estimate, inout float throughput, in float wavelength, in Material bounceMaterial) {
        float hr = 1.0 + InvC1(0.999);

        float U = RandNextF();

        vec3 wm = SampleD_wi(-direction, RandNext2F(), vec2(bounceMaterial.roughness));

        hr = SampleHeight(direction, vec2(bounceMaterial.roughness), hr, U);

        if(hr == 1e35) {
            return direction; //I think this is fine
        }

        direction = SampleOpaqueDielectricFacetBSDF(-direction, wm, estimate, throughput, wavelength, bounceMaterial);

        hr = SampleHeight(direction, vec2(bounceMaterial.roughness), hr, U);

        if(hr != 1e35) {
            throughput = 0.0;
        }

        if(hr != hr || direction.z != direction.z) {
            return vec3(0, 0, 1);
        }

        if(isnan(throughput)) {
            throughput = 0.0;
        }

        return direction;
    }

    float EstimateSingleScatteringBSDF(in vec3 incomingDirection, in vec3 outgoingDirection, in float wavelength, in Material bounceMaterial) {
        vec2 rng = RandNext2F();
        vec2 alpha = vec2(bounceMaterial.roughness);
        vec3 wm = SampleD_wi(incomingDirection, RandNext2F(), alpha);

        float Lambda_i = Smith_Lambda(incomingDirection, alpha);
        float Lambda_o = Smith_Lambda(outgoingDirection, alpha);

        float diffuse;
        {
            float G2_given_G1 = (1.0 + Lambda_i) / (1.0 + Lambda_i + Lambda_o);

            float energyConservationFactor = 1.0;

            if(bounceMaterial.thinFilm) {
                energyConservationFactor = HemisphericalAlbedoIntegration(bounceMaterial, wavelength);
            } else {
                energyConservationFactor -= HemisphericalAlbedo(bounceMaterial.n2.r / bounceMaterial.n1.r);
            }

            diffuse  = saturate(1.0 / (pi * max(0.0, dot(wm, outgoingDirection)) * G2_given_G1));
            diffuse *= bounceMaterial.albedo;
            //diffuse *= 1.0 - FresnelNonPolarized_R(dot(incomingDirection, wm), bounceMaterial.n1, bounceMaterial.n2);
            //diffuse *= 1.0 - FresnelNonPolarized_R(dot(outgoingDirection, wm), bounceMaterial.n1, bounceMaterial.n2);
            diffuse /= energyConservationFactor;
        }

        float specular;
        {
            vec3 halfway = normalize(incomingDirection + outgoingDirection);

            float G2 = 1.0 / (1.0 + Lambda_i + Lambda_o);

            float numerator = D_wi(incomingDirection, halfway, alpha) * FresnelNonPolarized_R(dot(incomingDirection, halfway), bounceMaterial.n1, bounceMaterial.n2) * G2;
            float denominator = 4.0 * incomingDirection.z * outgoingDirection.z;

            specular = numerator / denominator;
        }

        return bounceMaterial.isMetal ? specular : diffuse;
    }
#endif