float P22(in vec2 slope, in vec2 alpha) {
    if (any(equal(alpha, vec2(0.0)))) { return 0.0; }

    slope /= alpha;

    float tmp1 = slope.x*slope.x;
    float tmp2 = slope.y*slope.y;

    float value = 1.0 / (pi * alpha.x * alpha.y) * exp(-tmp1 - tmp2);
    return value;
}

float Smith_Lambda(in vec3 wi, in vec2 alpha) {
    if(wi.z > 0.9999) {
        return 0.0;
    }
    if(wi.z < -0.9999) {
        return -1.0;
    }

    float theta_I = fAcos(wi.z);
    float a = 1.0 / tan(theta_I) / Alpha_I(wi, alpha);

    float value = 0.5 * (erf(a) - 1.0) + 0.28209479177387814347 / a * exp(-a*a);

    return value;
}

float projectedArea(in vec3 wi, in vec2 alpha) {
    if(wi.z > 0.9990) {
        return 0.0;
    }
    if(wi.z < -0.9990) {
        return -1.0;
    }

    float alphaI = Alpha_I(wi, alpha);
    float theta_I = fAcos(wi.z);
    float a = 1.0 / tan(theta_I) / alphaI;

    float value = 0.5 * (erf(a) + 1.0) * wi.z + 0.28209479177387814347 * alphaI * sin(theta_I) * exp(-a*a);

    return value;
}

vec2 SampleP22_11(in float theta_I, in vec2 U) {
    vec2 slope;

    if(theta_I < 0.0001) {
        float r = sqrt(U.x/(1.0-U.x));
        float phi = tau * U.y;
        slope = vec2(r * cos(phi), r * sin(phi));

        return slope;
    }

    float sinTheta_I = sin(theta_I);
    float cosTheta_I = cos(theta_I);

    float slope_I = cosTheta_I/sinTheta_I;

    float a = slope_I;
    float projectedArea = 0.5 * (erf(a) + 1.0) * cosTheta_I + 0.28209479177387814347 * sinTheta_I * exp(-a*a);
    if(projectedArea < 0.0001 || projectedArea != projectedArea) {
        return vec2(0.0);
    }

    float c = 1.0 / projectedArea;

    float erf_min = -0.9999;
    float erf_max = max(erf_min, erf(slope_I));
    float erf_current = 0.5 * (erf_min + erf_max);

    while(erf_max-erf_min > 0.000001) {
        if(!(erf_current >= erf_min && erf_current <= erf_max)) {
            erf_current = 0.5 * (erf_min + erf_max);
        }

        float slope = erfinv(erf_current);

        float CDF = slope >= slope_I ? 1.0 : c * (0.28209479177387814347 * sinTheta_I * exp(-slope*slope) + cosTheta_I * (0.5 + 0.5 * erf(slope)));
        float diff = CDF - U.x;

        if(abs(diff) < 0.000001) {
            break;
        }

        if(diff > 0.0) {
            if(erf_max == erf_current) break;

            erf_max = erf_current;
        } else {
            if(erf_min == erf_current) break;

            erf_min = erf_current;
        }

        float derivative = 0.5 * c * cosTheta_I - 0.5 * sinTheta_I * slope;
        erf_current -= diff / derivative;
    }

    slope.x = erfinv(min(erf_max, max(erf_min, erf_current)));
    slope.y = erfinv(2.0 * U.y - 1.0);
    return slope;
}