#if !defined BRDF_HAPKE
#define BRDF_HAPKE
    /*
        https://serwiss.bib.hs-hannover.de/frontdoor/deliver/index/docId/2105/file/Bachelorarbeit_Alexander_Kuzminykh_20210818.pdf
    */

    float Hapke_1981(in vec3 normal, in vec3 incoming, in vec3 reflected, in float albedo) {
        const float h = 0.3;

        float g = acos(dot(incoming, reflected));
        float sinG = sin(g);
        float cosG = cos(g);
        float absG = abs(g);
        float tanABS_g = tan(absG);

        float cosI = saturate(dot(incoming, normal));
        float cosR = saturate(dot(reflected, normal));

        float B_g_ePart = exp(-h / tanABS_g);
        float B = 0.0;
        if(absG < pi / 2.0) {
            B = 1.9 * (1.0 - (tanABS_g / (2.0 * h)) * (3.0 - B_g_ePart) * (1.0 - B_g_ePart));
        } else {
            B = 0.0;
        }

        float gamma = sqrt(1.0 - albedo);
        float HR = (1.0 + 2.0 * cosR) / (1.0 + 2.0 * gamma * cosR);
        float HI = (1.0 + 2.0 * cosI) / (1.0 + 2.0 * gamma * cosI);

        float P = 0.8 * pi * (((sinG + (pi - g) * cosG) / pi) + (square(1.0 - cosG) / 10.0));

        float LS = albedo / (4.0 * pi * (cosR + cosI));

        return LS * ((1.0 + B) * P + HI * HR - 1.0);
    }

    float Hapke_2012(in vec3 normal, in vec3 incoming, in vec3 reflected, in float w) {
        const float b = 0.23955;
        const float c = 0.30452;
        const float hC = 0.7;
        const float hS = 0.2;
        const float BS0 = 1.8;
        const float BC0 = 1.1;
        const float thetaP = 24.3;
        const float phiT = 0.3;

        float cosG = dot(incoming, reflected);
        float cosI = saturate(dot(incoming, normal));
        float cosR = saturate(dot(reflected, normal));

        if(cosI < 0.0) {
            return 0.0;
        }

        float g = acos(cosG);
        float i = acos(cosI);
        float r = acos(cosR);

        float sinI = sin(i);
        float sinR = sin(r);

        float cosPsi = dot(normalize(incoming - cosI * normal), normalize(reflected - cosR * normal));
        cosPsi = clamp(cosPsi, -1.0, 1.0);
        float psi = acos(cosPsi);

        float psiHalf = psi / 2.0;
        float fPsi = exp(-2.0 * tan(psiHalf));
        float sinPsiHalf = sin(psiHalf);
        float sin2PsiHalf = square(sinPsiHalf);
        float psiPerPI = psi / pi;

        float tanThetaP = tan(thetaP * pi / 180.0);
        float tan2ThetaP = square(tanThetaP);

        float cotThetaP = 1.0 / tanThetaP;
        float cot2ThetaP = square(cotThetaP);

        float tanI = tan(i);
        float tanR = tan(r);

        float cotI = 1.0 / tanI;
        float cotR = 1.0 / tanR;

        float cot2R = square(cotR);
        float cot2I = square(cotI);

        float E_1_i = exp(-2.0 / pi * cotThetaP * cotI);
        float E_1_r = exp(-2.0 / pi * cotThetaP * cotR);
        float E_2_i = exp(-1.0 / pi * cot2ThetaP * cot2I);
        float E_2_r = exp(-1.0 / pi * cot2ThetaP * cot2R);

        float chiThetaP = 1.0 / sqrt(1.0 + pi * tan2ThetaP);
        float etaI = chiThetaP * (cosI + sinI * tanThetaP * E_2_i / (2.0 - E_1_i));
        float etaR = chiThetaP * (cosR + sinR * tanThetaP * E_2_r / (2.0 - E_1_r));

        float mR0 = chiThetaP;
        float mR = chiThetaP;

        float S;
        if(i <= r) {
            mR0 *= cosI + sinI * tanThetaP * (cosPsi * E_2_r + sin2PsiHalf * E_2_i) / (2.0 - E_1_r - psiPerPI * E_1_i);
            mR *= cosR + sinR * tanThetaP * (E_2_r - sin2PsiHalf * E_2_i) / (2.0 - E_1_r - psiPerPI * E_1_i);
            S = mR / etaR * mR0 / etaI * chiThetaP / (1.0 - fPsi + fPsi * chiThetaP * (mR0 / etaI));
        } else {
            mR0 *= cosI + sinI * tanThetaP * (E_2_i - sin2PsiHalf * E_2_r) / (2.0 - E_1_i - psiPerPI * E_1_r);
            mR *= cosR + sinR * tanThetaP * (cosPsi * E_2_i + sin2PsiHalf * E_2_r) / (2.0 - E_1_i - psiPerPI * E_1_r);
            S = mR / etaR * mR0 / etaI * chiThetaP / (1.0 - fPsi + fPsi * chiThetaP * (mR / etaR));
        }

        float kPhiTerm = 1.209 * pow(phiT, 2.0 / 3.0);
        float K = -log(1.0 - kPhiTerm) / kPhiTerm;

        float gHalf = g / 2.0;
        float tanGHalf = tan(gHalf);
        float tanGhalfPerhC = tanGHalf / hC;
        float BC = (1.0 + (1.0 - exp(-tanGhalfPerhC)) / tanGhalfPerhC) / (2.0 * square(1.0 + tanGhalfPerhC));
        if(isnan(BC)) {
            BC = 0.0;
        }
        if(isinf(BC)) {
            BC = 1.0;
        }

        float r0Term = sqrt(1.0 - w);
        float r0 = (1.0 - r0Term) / (1.0 + r0Term);

        float LS = mR0 / (mR0 + mR);

        float b2 = square(b);
        float oneMinusb2 = 1.0 - b2;
        float twoBcosG = 2.0 * b * cosG;

        float pG = (1.0 + c) / 2.0 * oneMinusb2 / pow(1.0 - twoBcosG + b2, 1.5) + (1.0 - c) / 2.0 * oneMinusb2 / pow(1.0 + twoBcosG + b2, 1.5);

        float BS = 1.0 / (1.0 + tanGHalf / hS);

        float xI = mR0 / K;
        float xR = mR / K;
        float HI = 1.0 / (1.0 - w * xI * (r0 + (1.0 - 2.0 * r0 * xI) / 2.0 * log((1.0 + xI) / xI)));
        float HR = 1.0 / (1.0 - w * xR * (r0 + (1.0 - 2.0 * r0 * xR) / 2.0 * log((1.0 + xR) / xR)));

        float M = HI * HR - 1.0;

        return (LS / cosI) * K * w / (4.0 * pi) * (pG * (1.0 + BS0 * BS) + M) * (1.0 + BC0 * BC) * S;
    }
#endif