#if !defined INDEX_OF_REFRACTION
#define INDEX_OF_REFRACTION
    #include "iorConstants.glsl"

    int BinarySearchIron(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (IronWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (IronWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchGold(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (GoldWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (GoldWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchAluminum(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (AluminumWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (AluminumWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchChromium(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (ChromiumWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (ChromiumWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchCopper(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (CopperWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (CopperWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchLead(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (LeadWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (LeadWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchPlatinum(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (PlatinumWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (PlatinumWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    int BinarySearchSilver(int lowIndex, int highIndex, float toFind) {
        while (lowIndex < highIndex) {
            int midIndex = (lowIndex + highIndex) >> 1;
            if (SilverWavelengths[midIndex] < toFind) {
                lowIndex = midIndex + 1;
            }
            else if (SilverWavelengths[midIndex] > toFind) {
                highIndex = midIndex;
            }
            else {
                return midIndex;
            }
        }
        return -1;
    }

    complexFloat IronIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, IronWavelengths[0]);
        float end = min(lambdaMax, IronWavelengths[43]);

        int idx = int(max(distance(0.0, float(BinarySearchIron(0, 43, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float ironN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 44 && end >= IronWavelengths[entry]; ++entry) {
                float a = IronWavelengths[entry],
                    b = IronWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = IronN[entry],
                    fb = IronN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            ironN = result / (lambdaMax - lambdaMin);
        }

        float ironK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 44 && end >= IronWavelengths[entry]; ++entry) {
                float a = IronWavelengths[entry],
                    b = IronWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = IronK[entry],
                    fb = IronK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            ironK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(ironN, ironK);
    }

    complexFloat GoldIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, GoldWavelengths[0]);
        float end = min(lambdaMax, GoldWavelengths[86]);

        int idx = int(max(distance(0.0, float(BinarySearchGold(0, 86, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float goldN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 87 && end >= GoldWavelengths[entry]; ++entry) {
                float a = GoldWavelengths[entry],
                    b = GoldWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = GoldN[entry],
                    fb = GoldN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            goldN = result / (lambdaMax - lambdaMin);
        }

        float goldK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 87 && end >= GoldWavelengths[entry]; ++entry) {
                float a = GoldWavelengths[entry],
                    b = GoldWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = GoldK[entry],
                    fb = GoldK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            goldK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(goldN, goldK);
    }

    complexFloat AluminumIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, AluminumWavelengths[0]);
        float end = min(lambdaMax, AluminumWavelengths[54]);

        int idx = int(max(distance(0.0, float(BinarySearchAluminum(0, 54, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float aluminumN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 55 && end >= AluminumWavelengths[entry]; ++entry) {
                float a = AluminumWavelengths[entry],
                    b = AluminumWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = AluminumN[entry],
                    fb = AluminumN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            aluminumN = result / (lambdaMax - lambdaMin);
        }

        float aluminumK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 55 && end >= AluminumWavelengths[entry]; ++entry) {
                float a = AluminumWavelengths[entry],
                    b = AluminumWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = AluminumK[entry],
                    fb = AluminumK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            aluminumK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(aluminumN, aluminumK);
    }

    complexFloat ChromiumIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, ChromiumWavelengths[0]);
        float end = min(lambdaMax, ChromiumWavelengths[43]);

        int idx = int(max(distance(0.0, float(BinarySearchChromium(0, 43, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float chromiumN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 44 && end >= ChromiumWavelengths[entry]; ++entry) {
                float a = ChromiumWavelengths[entry],
                    b = ChromiumWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = ChromiumN[entry],
                    fb = ChromiumN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            chromiumN = result / (lambdaMax - lambdaMin);
        }

        float chromiumK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 44 && end >= ChromiumWavelengths[entry]; ++entry) {
                float a = ChromiumWavelengths[entry],
                    b = ChromiumWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = ChromiumK[entry],
                    fb = ChromiumK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            chromiumK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(chromiumN, chromiumK);
    }

    complexFloat CopperIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, CopperWavelengths[0]);
        float end = min(lambdaMax, CopperWavelengths[76]);

        int idx = int(max(distance(0.0, float(BinarySearchCopper(0, 76, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float copperN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 77 && end >= CopperWavelengths[entry]; ++entry) {
                float a = CopperWavelengths[entry],
                    b = CopperWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = CopperN[entry],
                    fb = CopperN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            copperN = result / (lambdaMax - lambdaMin);
        }

        float copperK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 77 && end >= CopperWavelengths[entry]; ++entry) {
                float a = CopperWavelengths[entry],
                    b = CopperWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = CopperK[entry],
                    fb = CopperK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            copperK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(copperN, copperK);
    }

    complexFloat LeadIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, LeadWavelengths[0]);
        float end = min(lambdaMax, LeadWavelengths[146]);

        int idx = int(max(distance(0.0, float(BinarySearchLead(0, 146, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float leadN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 147 && end >= LeadWavelengths[entry]; ++entry) {
                float a = LeadWavelengths[entry],
                    b = LeadWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = LeadN[entry],
                    fb = LeadN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            leadN = result / (lambdaMax - lambdaMin);
        }

        float leadK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 147 && end >= LeadWavelengths[entry]; ++entry) {
                float a = LeadWavelengths[entry],
                    b = LeadWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = LeadK[entry],
                    fb = LeadK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            leadK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(leadN, leadK);
    }

    complexFloat PlatinumIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, PlatinumWavelengths[0]);
        float end = min(lambdaMax, PlatinumWavelengths[146]);

        int idx = int(max(distance(0.0, float(BinarySearchPlatinum(0, 146, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float platinumN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 147 && end >= PlatinumWavelengths[entry]; ++entry) {
                float a = PlatinumWavelengths[entry],
                    b = PlatinumWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = PlatinumN[entry],
                    fb = PlatinumN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            platinumN = result / (lambdaMax - lambdaMin);
        }

        float platinumK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 147 && end >= PlatinumWavelengths[entry]; ++entry) {
                float a = PlatinumWavelengths[entry],
                    b = PlatinumWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = PlatinumK[entry],
                    fb = PlatinumK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            platinumK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(platinumN, platinumK);
    }

    complexFloat SilverIOR(float wavelength) {
        float lambdaMin = wavelength - 0.5, lambdaMax = wavelength + 0.5;
        float start = max(lambdaMin, SilverWavelengths[0]);
        float end = min(lambdaMax, SilverWavelengths[146]);

        int idx = int(max(distance(0.0, float(BinarySearchSilver(0, 146, start))), 1.0) - 1.0);

        if (end <= start) {
            return complexFloat(0.0, 0.0);
        }

        float silverN = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 147 && end >= SilverWavelengths[entry]; ++entry) {
                float a = SilverWavelengths[entry],
                    b = SilverWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = SilverN[entry],
                    fb = SilverN[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            silverN = result / (lambdaMax - lambdaMin);
        }

        float silverK = 0.0;
        {
            float result = 0.0;
            int entry = idx;
            for (; entry < 147 && end >= SilverWavelengths[entry]; ++entry) {
                float a = SilverWavelengths[entry],
                    b = SilverWavelengths[entry + 1],
                    ca = max(a, start),
                    cb = min(b, end),
                    fa = SilverK[entry],
                    fb = SilverK[entry + 1],
                    invAB = 1.0 / (b - a);

                if (ca >= cb) {
                    continue;
                }

                float interpA = mix(fa, fb, (ca - a) * invAB);
                float interpB = mix(fa, fb, (cb - a) * invAB);

                result += 0.5 * (interpA + interpB) * (cb - ca);
            }

            silverK = result / (lambdaMax - lambdaMin);
        }
        return complexFloat(silverN, silverK);
    }
#endif