uniform float aspectRatio;

struct camera {
    float scale;
    float sensorWidth;
    float sensorHeight;
    float r;
    float fstops;
    float shutterTime;
    float ISO;

    vec2 sensorSize;
    vec2 focalLength;
    vec2 apertureRadius;
}cm;

vec2 distort(vec2 p) {
    float theta  = atan(p.y, p.x);
    float radius = length(p);
    radius = pow(radius, 5.0);
    p.x = radius * cos(theta);
    p.y = radius * sin(theta);
    return 0.5 * (p + 1.0);
}

vec2 BokehShape() {
    const int blades = 3;

    vec2 uv = RandNext2F();

    vec2 axis;
    if(blades > 0) {
        const float angle = tau/blades;

        uv.x *= blades;
        float blade = floor(uv.x);
        uv.x = fract(uv.x);

        mat2 rot = Rotate((blade+0.0) * angle);

        axis  = rot * vec2(cos(angle / 2.0), sin(angle / 2.0) * (uv.x * 2.0 - 1.0));
        axis *= 1.0 - pow(1.0 - sqrt(uv.y), 1.5);
    } else {
        axis  = vec2(0.0);
        axis  = vec2(cos(tau * uv.x), sin(tau * uv.x));
        axis *= 1.0 - pow(1.0 - sqrt(uv.y), 1.0);
    }
    vec2 anamorphic = vec2(1.0 * ASPECT_RATIO, 1.0 * rcp(ASPECT_RATIO));

    return axis * anamorphic;
}

void DefineCamera() {
    cm.sensorWidth = CAMERA_SCALE * 1e-3 * 36;
    cm.sensorHeight = CAMERA_SCALE * 1e-3 * 36;
    #ifdef LEGACY_LENS_SYSTEM
        cm.sensorWidth = CAMERA_SCALE * 1e-3 * 36;
        cm.sensorHeight = CAMERA_SCALE * 1e-3 * 27;
    #endif
    cm.r = length(BokehShape());
    cm.fstops = F_STOPS;
    cm.shutterTime = 1.0/SHUTTER_SPEED;
    cm.ISO = SENSITIVITY_TO_LIGHT;

    cm.sensorSize = vec2(cm.sensorWidth, cm.sensorHeight);
    cm.focalLength  = 0.5 * cm.sensorSize * vec2(gbufferProjection[0].x, gbufferProjection[1].y);
    #ifdef MANUAL_CAMERA
        cm.focalLength  = 0.5 * cm.sensorSize / (2.0 * vec2(aspectRatio, 1.0) * tan(radians(FOV) / 2.0));
    #endif
    cm.focalLength *= 1.0 + cm.r * 0.0;
    cm.apertureRadius = (cm.focalLength / cm.fstops) / 2.0;
}