/* RENDERTARGETS:9 */
vec4 color;
vec4 purkinjeEffect;

layout (rgba32f) uniform image2D colorimg0;
layout (rgba32f) uniform image2D colorimg5;

uniform sampler2D colortex0;
uniform sampler2D colortex5;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D colortex8;
uniform sampler2D colortex13;
uniform sampler2D colortex15;

uniform sampler2D noisetex;

uniform sampler2D colortex10;

uniform sampler2D shadowcolor0;
uniform usampler2D shadowcolor1;

uniform mat4 gbufferModelViewInverse, gbufferProjectionInverse;
uniform mat4 gbufferProjection, gbufferModelView;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;

uniform vec3 cameraPosition, previousCameraPosition;
//uniform vec3 sunVector;

uniform float viewWidth, viewHeight;

uniform int worldTime;
uniform int worldDay;

uniform int frameCounter;

layout(location = 0) in vec2 textureCoordinate;

const bool colortex1MipmapEnabled = true;

const bool colortex0Clear = false;
const bool colortex3Clear = false;
const bool colortex5Clear = false;
const bool colortex6Clear = false;
const bool colortex7Clear = false;
const bool colortex8Clear = false;


/*
    const vec4 shadowcolor1ClearColor = vec4(-170141183460469231731687303715884105728, 0, 0, 0); 
*/

vec3 sunVector, moonVector;
vec2 viewSize;
float time;

#include "/lib/utility/universal.glsl"

#include "/lib/rng/msws.glsl"
#include "/lib/rng/bayer.glsl"

#include "/lib/voxel/voxelization.glsl"
#include "/lib/voxel/voxelBasis.glsl"

#include "/lib/fragment/camera/cameraSystem.glsl"
#include "/lib/fragment/camera/airy.glsl"

#include "/lib/utility/sunMoonCalc.glsl"

float BinarySearch_CIE(float lowBound, float highBound, float toFind, in vec3 weights) {
    float midIndex = (lowBound + highBound) / 2.0;

    int lookupRes = textureSize(colortex11, 0).x * 2;
    for (int x = 0; x < log2(lookupRes); ++x) {
        float coordinate = midIndex;
        float coordTweaked = (coordinate * float(lookupRes - 1) + 0.5) / lookupRes;

        float value = saturate(weights.x * texture(colortex11, vec2(coordTweaked, 0)).x + weights.y * texture(colortex11, vec2(coordTweaked, 0)).y + weights.z * texture(colortex11, vec2(coordTweaked, 0)).z); //The CDF value should never go above 1.

        if (value < toFind) {
            lowBound = midIndex;
        }
        else if (value > toFind) {
            highBound = midIndex;
        }
        else {
            return midIndex;
        }

        midIndex = (lowBound + highBound) / 2.0;
    }
    return midIndex;
}

#include "/lib/fragment/path/estimate.glsl"

vec2 reproject(vec2 coord, float depth) {
    vec4 position	= vec4(coord, depth, 1.0)*2.0-1.0;
        position    = gbufferProjectionInverse*position;
        position   /= position.w;
        position    = gbufferModelViewInverse*position;

		position   += vec4(cameraPosition-previousCameraPosition, 0.0);
		position    = gbufferPreviousModelView*position;
		position    = gbufferPreviousProjection*position;

    return (position.xy/position.w)*0.5+0.5;
}

float BetaM() {
    float mie = 2.6650023e14;
    float c = (0.2 * TURPIDITY) * 10e-19;
    return 0.434 * c * mie;
}

float BetaM(in float wl) {
    float unMix = (wl - 390.0) / 441.0;
    float mie = 2.6650023e14;
    float c = (0.2 * TURPIDITY) * 10e-19;
    return mix(0.3, 0.4, unMix) * c * mie;
}

float RayleighCoefficient(float radius, float lambda, complexFloat ior) {
    //Calculate the Rayleigh coefficient using the Rayleigh approximation of Mie Theory.
	float x = 2.0 * pi * radius / lambda;

	complexFloat ior4 = complexMul(complexMul(ior, ior), complexMul(ior, ior));

	complexFloat a1 = complexMul(complexMul(2.0 * cube(x) / 3.0, complexDiv(complexSub(complexMul(ior, ior), 1.0), complexAdd(complexMul(ior, ior), 2.0))), complexFloat(0.0, 1.0));
				 a1 = complexAdd(
					 a1,
					 complexMul(
						 (2.0 * (x * x * x * x * x)) / 5.0,
						 complexMul(
							 complexDiv(
							 	 complexMul(
							 	 	 complexSub(complexSquare(ior), 2.0),
							 	 	 complexAdd(complexSquare(ior), 1.0)
							 	 ),
							 	 complexMul(
							 	 	 complexAdd(complexSquare(ior), 2.0),
							 	 	 complexAdd(complexSquare(ior), 2.0)
							 	 )
							 ),
							 complexFloat(0.0, 1.0)
						 )
					 )
				 );

	complexFloat ratio = complexDiv(complexSub(complexSquare(ior), 1.0), complexAdd(complexSquare(ior), 2.0));
	complexFloat extinction = complexMul(
		4.0 * x, 
		complexMul(
			ratio, 
			complexAdd(
				1.0,
				complexMul(
					square(x) / 15.0, 
					complexMul(
						ratio, 
						complexDiv(
							complexAdd(
								ior4,
								complexAdd(
									complexMul(
										27.0,
										complexSquare(ior)
									), 
									38.0
								)
							),
							complexAdd(
								complexMul(2.0, complexSquare(ior)),
								3.0
							)
						)
					)
				)
			)
		)
	);

	float Qsca = 8.0 / 3.0 * pow(x, 4.0) * square(complexAbs(ratio));
	float Qext = abs(extinction.i + Qsca);

    return Qsca;
}

float BetaR(in float wl) {
    //Two different methods of calculating the Rayleigh coefficient.
    float nanometers = wl * 1e-9;
    float micrometers = wl * 1e-6;

    float F_N2 = 1.034 + 3.17e-4 * (1.0 / pow(wl, 2.0));
    float F_O2 = 1.096 + 1.385e-3 * (1.0 / pow(wl, 2.0)) + 1.448e-4 * (1.0 / pow(wl, 4.0));
    float CCO2 = 0.045;
    float kingFactor = (78.084 * F_N2 + 20.946 * F_O2 + 0.934 + CCO2 * 1.15) / (78.084 + 20.946 + 0.934 + CCO2);
    float n = square(Air(wl * 1e-3)) - 1.0;

    float cheap = ((8.0 * pow(pi, 3.0) * pow(n, 2.0)) / (3.0 * airNumberDensity * pow(nanometers, 4.0))) * kingFactor;

    return RayleighCoefficient(1e-9, nanometers, complexFloat(Air(wl * 1e-3), 0.0)) * 3.0 * kingFactor * (airNumberDensity / 1e16);
}

float SpectralOzone(float wl) {
    if(wl < 390.0 || wl > 830.0) return 0.0;
    wl = (wl - 390.0) / 1.0;

    return 0.0001 * ozoneNumberDensity * ozoneCrossSection[int(wl)];
}

float HitDistance() {
    vec2 aa = (2.0 * RandNext2F() - 1.0) * rcp(viewSize);

    vec2 coordinate = vec2(0.5);

    vec2 xy = coordinate * 2.0 - 1.0;
         xy += aa;
         xy *= cm.sensorSize / (2.0 * cm.focalLength);
    vec3 viewVector = normalize(vec3(xy, -1.0));
         viewVector = mat3(gbufferModelViewInverse) * viewVector;

    vec3 hitPosition = VoxelSpaceFromSceneSpace(mat3(gbufferModelViewInverse) * vec3(0.0) + gbufferModelViewInverse[3].xyz);
    ivec3 hitIndex = ivec3(hitPosition);
    vec3 rayDirection = viewVector;

    vec2 uv;
    mat3 tbn;

    float hitDist;

    VoxelData data;
    vec2 voxelmapCoord;
    bool hit = RaytraceVoxel(hitPosition, hitIndex, rayDirection, hitPosition, hitIndex, hitDist, data);

    return clamp(distance(SceneSpaceFromVoxelSpace(hitPosition), gbufferModelViewInverse[3].xyz), 0.0, 5e3);
}

void EnableDOF(inout vec3 hitPosition, inout vec3 viewVector, in vec2 dofShape, in float wavelength) {
    hitPosition.xy += cm.apertureRadius * dofShape;

    float focus = HitDistance() * (1.0 - square(cm.r) * 0.0);

    const float tiltTheta = radians(0.0);
    const float tiltPhi = radians(0.0);
    const vec3 n = vec3(vec2(-sin(tiltPhi), cos(tiltPhi)) * sin(2.0 * tiltTheta), cos(2.0 * tiltTheta));
    vec3 p = vec3(0.0, 0.0, focus);

    float focusDistance = dot(p, n) / abs(dot(viewVector, n));
          //focusDistance = focusDistance * (1.0 + (wavelength - 550.0) / 100.0);

    vec3 focusPoint = viewVector * focusDistance;

    viewVector = focusPoint - hitPosition;
    viewVector = normalize(viewVector);
}

vec3 FisheyeProjection(in float cx, in float cy, in float fov) {
    vec2 coordinate = vec2(cx, cy);
    vec2 aspect_ratio = vec2(viewSize.x / viewSize.y, 1);
	vec3 p = vec3((2.0 * coordinate - 1.0) * aspect_ratio * fov, -1.0);

	float z2 = p.x * p.x + p.y * p.y;

    if (z2>2.0) {
        return vec3(-1.0);
    }

	float phi = atan(p.y, p.x);
	float theta = acos(1.0 - z2);

    vec3 dir = vec3(sin(theta) * cos(phi), cos(theta), sin(theta) * sin(phi));

    return dir;
}

float LensMakersEquation(in float n, in float R1, in float R2, in float d) {
    return (n - 1.0) * ((1.0 / R1) - (1.0 / R2) + (((n - 1.0) * d) / (n * R1 * R2)));
}

void LensIntersect(inout vec3 lensPosition, inout vec3 rayDirection, inout vec3 refractDirection, out vec3 tmpPosition, out float focalLength, in float radius1, in float radius2, in float d, in float nT, in float nI) {
    float lensThickness = d;
    float lensThicknessHalf = lensThickness / 2.0;

    focalLength = (1.0 / LensMakersEquation(nT, radius1, radius2, lensThickness));

    float radius = radius1 * sin(acos(0.98));
    float x = cos(asin(radius / radius1));

	// I changed Y from 0.0 to a non-zero number and it seems to have fixed half the lens being black.
	// Without understanding most of this code, I think it's either a faux-precision error, or something-
	// like the lens ever-so-slightly intersecting the sensor.
    vec3 position = vec3(0.0, 1e-5, (focalLength - radius1) + lensThicknessHalf);
    lensPosition = GenerateUnitVector(vec2(RandNextF(), RandNextF() * (1.0 - x) + x)) * radius1 + position;
    vec3 lensNormal   = normalize(lensPosition - position);
    rayDirection = GenerateCosineVector(lensNormal, RandNext2F());
    bool inside = dot(lensNormal, rayDirection) > 0.0;

    complexFloat n1 = complexFloat(nI, 0.0), n2 = complexFloat(nT, 0.0);
    if(inside) {
        n1 = complexFloat(nT, 0.0);
        n2 = complexFloat(nI, 0.0);
    }
    float fresnel_R = FresnelNonPolarized_R(dot(lensNormal, -rayDirection), n1, n2);
    bool specularBounce = fresnel_R > RandNextF();

    refractDirection = refractIT(-rayDirection, lensNormal, nI, nT);

    tmpPosition = lensPosition - position;
    {
        vec3 position = vec3(0.0, 0.0, (focalLength - radius2) - lensThicknessHalf);

        float dist;
        if(IntersectSphere(lensPosition - position, refractDirection, radius2, dist)) {
            vec3 oldRayDirection = refractDirection;
            lensPosition += dist * refractDirection;
            vec3 lensNormal = normalize(lensPosition - position);
            bool inside = dot(lensNormal, rayDirection) > 0.0;

            complexFloat n1 = complexFloat(nI, 0.0), n2 = complexFloat(nT, 0.0);
            if(inside) {
                n1 = complexFloat(nT, 0.0);
                n2 = complexFloat(nI, 0.0);
            }
            float fresnel_R = FresnelNonPolarized_R(dot(lensNormal, oldRayDirection), n1, n2);
            bool specularBounce = fresnel_R > RandNextF();

            refractDirection = refractIT(refractDirection, lensNormal, nI, nT);
        }
    }
}

ivec2 pixelPre;

void main() {
    time = (worldDay * 24000.0 + (worldTime + 300.0) * 50.0 * 72.0);
    viewSize = vec2(viewWidth, viewHeight);

	uint seed = uint(gl_FragCoord.x * viewSize.y + gl_FragCoord.y);
	     seed = seed * 720720u + uint(frameCounter);

    init_msws(seed);

    float size = viewWidth / 10.0;
    vec2 position = floor(gl_FragCoord.xy / size);

    vec3 hitPosition;

    float purkinje;
    vec3 rgb;

    vec2 pixelOnSensor;

    float wavelength;
    if(mod(position.x + mod(position.y + float(frameCounter % 2), 2.0), 2.0) == 0.0) {
        vec3 sunPosition  =  SunPosition(time, 43.9, 86.0);
        vec3 moonPosition = MoonPosition(time, 43.9, 86.0);

        moonAngularRadius = moonRadius / (moonPosition.z * 1e3);

        DefineCamera();

        sunVector  =      normalize(vec3(sin(sunPosition.y) * cos(sunPosition.x), cos(sunPosition.y) * cos(sunPosition.x), sin(sunPosition.x)));
        moonVector = normalize(vec3(sin(moonPosition.y) * cos(moonPosition.x), cos(moonPosition.y) * cos(moonPosition.x), sin(moonPosition.x)));

        float wavelengthUniform = (441.0 * RandNextF()) + 390.0;
        float n = Glass(wavelengthUniform * 1e-3);
        float n_Air = Air(wavelengthUniform * 1e-3);

        float radius1 = LENS_OUTER_RADIUS;
        float radius2 = LENS_INNER_RADIUS; 
        float d = LENS_THICKNESS;

        float sensorDistance = DISTANCE_TO_SENSOR;

        vec3 history;
        {
            vec3 lensPosition;
            vec3 rayDirection;
            vec3 refractDirection;
            vec3 tmpPosition;
            vec3 lensNormal;
            float focalLength;
            LensIntersect(lensPosition, rayDirection, refractDirection, tmpPosition, focalLength, radius1, radius2, d, n, n_Air);

            float t = 0.0;
            int sensor = RayPlaneIntersection(
                lensPosition + vec3(0, 0, 1.0), refractDirection,
                vec3(0.0, 0.0, 0.0), vec3(0.0, 0.0, 1.0),
                t
            );
            vec3 sensorPosition = lensPosition + t * refractDirection;
            vec2 pixelOnSensor = sensorPosition.xy;
                 pixelOnSensor = (pixelOnSensor * viewSize.x * rcp(2.0)) + (viewSize / 2.0);

            ivec2 pixel = ivec2(gl_FragCoord.st);
            #ifndef LEGACY_LENS_SYSTEM
                pixel = ivec2(pixelOnSensor);
            #endif

            vec3 history = texelFetch(colortex0, pixel, 0).rgb;
            int weight = 0;
            for(int x = -6; x <= 6; ++x) {
                for(int y = -6; y <= 6; ++y) {
                    history += texelFetch(colortex0, pixel + ivec2(x, y) * 2, 0).rgb;
                }
            }
        }

        vec3 CIE_CMF_WEIGHTS = vec3(WAVELENGTH_SAMPLING_X, WAVELENGTH_SAMPLING_Y, WAVELENGTH_SAMPLING_Z) / dot(vec3(WAVELENGTH_SAMPLING_X, WAVELENGTH_SAMPLING_Y, WAVELENGTH_SAMPLING_Z), vec3(1.0));

        float catch = history == vec3(0.0) ? 1.0 : 0.2;
        CIE_CMF_WEIGHTS = mix(saturate(history / dot(vec3(1.0), history)), CIE_CMF_WEIGHTS, catch);

        float wUniform = WAVELENGTH_SAMPLING_UNIFORM;

        scale = 683.0 / texelFetch(colortex11, ivec2(int(2.99792458e17 / 540e12) - 390, 1), 0).y;

        wavelength;
        if (RandNextF() < wUniform) {
            wavelength = wavelengthUniform;
        } else {
            wavelength = (441.0 * BinarySearch_CIE(0.0, 1.0, RandNextF(), CIE_CMF_WEIGHTS)) + 390.0;
        }

        vec3 cie = texture(colortex11, vec2(int(wavelength - 390.0) / 441.0, 1)).xyz;

        float wavelength_pdf_sampled = CIE_CMF_WEIGHTS.x * (cie.x / 113.042) + CIE_CMF_WEIGHTS.y * (cie.y / 113.042) + CIE_CMF_WEIGHTS.z * (cie.z / 113.042);
        wavelength_pdf_sampled = mix(wavelength_pdf_sampled, 1.0 / 441.0, wUniform);
        if(isinf(wavelength_pdf_sampled) || isnan(wavelength_pdf_sampled)) {
            wavelength_pdf_sampled = 1.0;
        }

        #ifdef LEGACY_LENS_SYSTEM
            float k = 2.0 * pi * 1.450 / wavelengthUniform;

            vec2 dofShape = BokehShape() * (2.0 * pi / (wavelengthUniform / 415.5)) * 0.0;

            vec2 aa = (2.0 * RandNext2F() - 1.0) * rcp(viewSize);

            vec2 uv  = (gl_FragCoord.st * rcp(viewSize)) * 2.0 - 1.0;
                uv += aa;
                uv *= cm.sensorSize / (2.0 * cm.focalLength);

            vec3 viewDirection = vec3(uv, -1.0);
            hitPosition = vec3(0.0);

            EnableDOF(hitPosition, viewDirection, dofShape, wavelengthUniform);

            float lensFresnel = 1.0;
        #else
            vec3 lensPosition;
            vec3 rayDirection;
            vec3 refractDirection;
            vec3 tmpPosition;
            vec3 lensNormal;
            float focalLength;
            LensIntersect(lensPosition, rayDirection, refractDirection, tmpPosition, focalLength, radius1, radius2, d, n, n_Air);

            float t = 0.0;
            int sensor = RayPlaneIntersection(
                lensPosition + vec3(0, 0, sensorDistance), refractDirection,
                vec3(0.0, 0.0, 0.0), vec3(0.0, 0.0, 1.0),
                t
            );
            vec3 sensorPosition = lensPosition + t * refractDirection;
            pixelOnSensor = sensorPosition.xy;
            pixelOnSensor = (pixelOnSensor * viewSize.x * rcp(2.0)) + (viewSize * rcp(2.0));

            vec3 viewDirection = vec3(rayDirection.x, rayDirection.y, -rayDirection.z);
            hitPosition = tmpPosition;

            float lensFresnel = FresnelNonPolarized_R(dot(lensNormal, rayDirection), complexFloat(n_Air, 0.0), complexFloat(n, 0.0));
        #endif

        hitPosition = VoxelSpaceFromSceneSpace(ViewSpaceToSceneSpace(hitPosition, gbufferModelViewInverse));
        viewDirection = normalize(mat3(gbufferModelViewInverse) * viewDirection);

        float rayleigh = BetaR(wavelength);
        float mie = BetaM(wavelength);
        float ozone = SpectralOzone(wavelength);

        vec3 attenuationCoefficients = vec3(rayleigh, mie / aerosolScatteringAlbedo, ozone);

        rgb = EstimatePathLighting(hitPosition, viewDirection, ivec3(floor(hitPosition)), attenuationCoefficients, lensFresnel, wavelength, wavelength_pdf_sampled, purkinje);
    }

    vec2 reprojected = reproject(textureCoordinate, ScreenSpaceToViewSpace(distance(SceneSpaceFromVoxelSpace(hitPosition), gbufferModelViewInverse[0].xyz), gbufferProjectionInverse));

    bool    moved  = gbufferProjection != gbufferPreviousProjection;
    moved = moved || gbufferModelView  != gbufferPreviousModelView;
    moved = moved || cameraPosition    != previousCameraPosition;

    ivec2 pixel = ivec2(gl_FragCoord.st);
    #ifndef LEGACY_LENS_SYSTEM
        pixel = ivec2(pixelOnSensor);
    #endif
    pixelPre = pixel;

    int frames = int(texelFetch(colortex0, pixel, 0).a);
    if(moved) {
        frames = 0;
    }

    //frames = clamp(frames, 0, 100);

    color.rgb = mix(texelFetch(colortex0, pixel, 0).rgb, rgb, 1.0 / float(++frames));
    purkinjeEffect.rgb = mix(texelFetch(colortex5, pixel, 0).rgb, vec3(purkinje, texelFetch(colortex5, pixel, 0).gb), 1.0 / float(frames));
    color.a   = frames;

    purkinjeEffect.gba = mix(texelFetch(colortex5, pixel, 0).gba, SpectrumToXYZ(Scotopic(wavelength), wavelength) * xyzToRGBMatrix, 1.0 / float(frames));

    imageStore(colorimg0, pixel, color);
    imageStore(colorimg5, pixel, purkinjeEffect);
}