#if !defined PATH_CELESTIAL
#define PATH_CELESTIAL
    const float sunRadius        = 6.95e8;
    const float sunDistance      = 1.49e11;
    const float sunAngularRadius = (sunRadius / sunDistance) * 1.0;

    const float moonRadius         = 1.7e6;
    float moonAngularRadius;

    float PhysicalSun(in vec3 sceneDirection, in vec3 lightDirection, in float a, in float luminance) {
        float cosTheta = dot(sceneDirection, lightDirection);
        float centerToEdge = clamp(acos(cosTheta) / sunAngularRadius, 0.0, 1.0);
        float cosSunRadius = cos(sunAngularRadius);
        if(cosTheta < cosSunRadius) return 0.0;

        float mu = sqrt(1.0 - (centerToEdge*centerToEdge*centerToEdge*centerToEdge*centerToEdge));
        float factor = pow(mu, a);
        float finalLuminance = luminance * factor;

        return finalLuminance;
    }

    vec2 EquirectangularCoord(in vec3 rayDirection) {
        float lon = atan(rayDirection.z, rayDirection.x);
        if(rayDirection.z < 0) {
            lon = 2 * pi - atan(-rayDirection.z, rayDirection.x);
        }

        float lat = acos(rayDirection.y);

        const vec2 rads = vec2(1.0 / (pi * 2.0), 1.0 / pi);
        vec2 sphereCoords = vec2(lon, lat) * rads;

        return sphereCoords;
    }

    float PhysicalMoon(in vec3 sceneDirection, in vec3 moonDirection, in vec3 sunDirection, in float albedo, in float shine, in float irradiance) {
        //Something broke with this function
        vec2 dists = RSI(-moonDirection, sceneDirection, moonAngularRadius);

        vec3 flatNormal = normalize(sceneDirection*dists.x - moonDirection);

        mat3 tbnMatrix = GetRotationMatrix(vec3(0.0, 0.0, 1.0), flatNormal);

        albedo = albedo;

        vec3 normal = flatNormal;

        float brdf = Hapke_2012(normal, sunDirection, -sceneDirection, albedo) * saturate(dot(normal, sunDirection));
        #ifdef USE_OLD_HAPKE
            brdf = Hapke_1981(normal, sunDirection, -sceneDirection, albedo) * saturate(dot(normal, sunDirection));
        #endif

        if(isnan(brdf)) {
            brdf = 0.0;
        }
        if(isinf(brdf)) {
            brdf = 1.0;
        }

        return dists.y > 0.0 ? albedo * saturate(dot(normal, sunDirection)) * irradiance / pi : 0.0;
    }
#endif