#if !defined PATH_ESTIMATE
#define PATH_ESTIMATE
    #include "/lib/fragment/bsdf/fresnel.glsl"

    #include "materials.glsl"

    #include "atmos.glsl"

    #include "gooch.glsl"

    struct Volume {
        float sigmaT;
        float sigmaS;
        float albedo;

        float forwardG;
        float backwardG;
    };

    Volume Empty = {0.0, 0.0, 0.0, 0.0, 0.0};

    bool FindNextInteractionMaterial(in vec3 rayPosition, in vec3 rayDirection, inout float volumeDist, in float distanceToTrace, in float sigmaT) {
        float t = -log(RandNextF()) / sigmaT;
        if(t > distanceToTrace) {
            return false;
        }

        volumeDist = t;

        return true;
    }

    void RandomWalkRemap(in float albedo, in float radius, inout float sigmaT, inout float sigmaS) {
        float a = 1.0 - exp(-(5.09406 + albedo + 2.61188 * square(albedo) - 4.31805 * cube(albedo)));
        float s = 1.9 - albedo + 3.5 * square(albedo - 0.8);

        sigmaT = 1.0 / max(radius * s, 1.0 / 3e38);
        sigmaS = sigmaT * a * albedo;
    }

    Volume DefineSubsurface(in float albedo, in float radius, in float wavelength, in int id) {
        Volume subsurface;

        subsurface.sigmaT;
        subsurface.sigmaS;

        if(id == 161) {
            albedo = albedo * (2.0 - albedo);
        }

        RandomWalkRemap(albedo, radius, subsurface.sigmaT, subsurface.sigmaS);

        if(id == 161) {
            subsurface.sigmaT = subsurface.sigmaT - subsurface.sigmaS;
            subsurface.sigmaS *= 0.0;
        }
        if(id == 157) {
            subsurface.sigmaT = subsurface.sigmaT - subsurface.sigmaS;
            subsurface.sigmaS *= 0.3;
            subsurface.sigmaT = subsurface.sigmaT + subsurface.sigmaS;
        }
        if(id == 158) {
            //subsurface.sigmaS *= 0.3;
            subsurface.sigmaT = subsurface.sigmaT - subsurface.sigmaS;
            subsurface.sigmaS *= 0.3;
            subsurface.sigmaT = subsurface.sigmaT + subsurface.sigmaS;
        }

        subsurface.albedo = saturate(subsurface.sigmaS / subsurface.sigmaT);

        return subsurface;
    }

    #include "/lib/fragment/bsdf/hapke.glsl"
    #include "celestial.glsl"

    #include "/lib/fragment/bsdf/bsdfCalculation/bsdf.glsl"
    #include "/lib/fragment/bsdf/bsdfCalculation/chronos.glsl"

    float SubsurfacePhase(in float cosTheta) {
        return mix(KleinNishina(cosTheta, 0.78), HenyeyGreensteinPhase(cosTheta, -0.75), 0.3);
    }
    float SkinPhase(in float cosTheta) {
        return mix(KleinNishina(cosTheta, 0.64), HenyeyGreensteinPhase(cosTheta, -0.75), 0.3);
    }

    vec3 SampleSubsurfacePhase(in float forwardG, in float backwardG) {
        if(RandNextF() < 0.3) {
            return SampleHenyeyGreenstein(RandNextF(), backwardG);
        }
        return SampleKleinNishina(forwardG);
    }

    #include "scotopic.glsl"

    float cot(in float x) {
        return cos(x) / sin(x);
    }

    uniform vec3 fogColor;

    vec3 EstimatePathLighting(in vec3 rayPosition, in vec3 rayDirection, in ivec3 hitIndex, in vec3 baseAttenuationCoefficients, in float lensFresnel, in float wavelength, in float wavelengthPDF, out float purkinje) {
        int component;

        float sunRadiance   = Plancks(5778.0, wavelength);
        float sunIrradiance = sunRadiance * coneAngleToSolidAngle(sunAngularRadius);

        float averageAlbedo = 0.072 * mix(1.0, 2.0, (wavelength - 390.0) / 441.0);

        float moonAlbedo = mix(1.0, 2.0, (wavelength - 390.0) / 441.0);

        float d = ToDays(time);
        vec2 cS = SunCoords(d);
        vec3 cM = MoonCoords(d);

        float phiT = acos(sin(cS.x) * sin(cM.x) + cos(cS.x) * cos(cM.x) * cos(cS.y - cM.y)),
        inc = atan((sunDistance / 1e3) * sin(phi), cM.z - (sunDistance / 1e3) * cos(phi)),
        angle = atan(cos(cS.x) * sin(cS.y - cM.y), sin(cS.x) * cos(cM.x) - cos(cS.x) * sin(cM.x) * cos(cS.y - cM.y));

        angle = angle - parallacticAngle;

        float fraction = (1.0 + cos(inc)) / 2.0;
        float P = 0.5 + 0.5 * inc * (angle < 0 ? -1 : 1) / pi;
        float phaseAngle = acos(2.0 * P / (100.0 - 1.0));

        float earthShine = 0.19 * 0.5 * (1.0 - sin((tau - phaseAngle) / 2.0) * tan((tau - phaseAngle) / 2.0) * log(cot((tau - phaseAngle) / 2.0)));
        float moonIrradiance = ((2.0 * moonAlbedo * square(moonRadius)) / (3.0 * square(cM.z * 1e3))) * (sunIrradiance * (1.0 - sin(phaseAngle / 2.0) * tan(phaseAngle / 2.0) * log(cot(phaseAngle / 4.0))));

        float estimate = 0.0;
        float throughput = 1.0 * lensFresnel;

        vec3 viewRayDirection = rayDirection;
        vec3 viewRayPosition  = rayPosition;

        const float moonPathProbability = 0.5;

        Volume subsurface = Empty;
        Volume airVolume = Empty;

        vec3 netherColor = SrgbToLinear(fogColor) * 0.9 + 0.1;

        float netherSky = 0.0;
        RGBToSpectrum(netherSky, wavelength, netherColor.r, netherColor.g, netherColor.b, 0);

        RandomWalkRemap(netherSky, 10.0, airVolume.sigmaT, airVolume.sigmaS);
        airVolume.albedo = saturate(airVolume.sigmaS / airVolume.sigmaT);

        #if DISPLAY == 0
            int pathSegment = 0;
            bool volumeBounce = false;
            bool isRefractRay = false;
            while(pathSegment < MAX_PATH_SEGMENTS) {
                #if WORLD == 0
                    #include "overworld.glsl"
                #elif WORLD == 1
                    #include "end.glsl"
                #elif WORLD == 2
                    #include "nether.glsl"
                #endif

                ++pathSegment;
            }

            if(isnan(estimate)) {
                estimate = 0.0;
            }

            //Accurate scotopic luminance.
            purkinje = estimate * Scotopic(wavelength);
            purkinje /= wavelengthPDF;
            purkinje *= scale;

            estimate /= wavelengthPDF;

            #ifndef EXACT_CIE_SAMPLING
                return SpectrumToXYZFit_CIE2012_Tech(estimate, wavelength) * xyzToRGBMatrix;
            #else
                return SpectrumToXYZ(estimate, wavelength) * xyzToRGBMatrix;
            #endif
        #elif DISPLAY == 1
            VoxelData data;
            vec2 voxelmapCoord;
            float dist;
            ivec3 viewHitIndex = ivec3(floor(viewRayPosition));
            bool hit = RaytraceVoxel(viewRayPosition, viewHitIndex, viewRayDirection, viewRayPosition, viewHitIndex, dist, data);

            return dot(data.normalMap, viewRayDirection) * vec3(1.0);
        #elif DISPLAY == 2
            VoxelData data;
            vec2 voxelmapCoord;
            float dist;
            ivec3 viewHitIndex = ivec3(floor(viewRayPosition));
            bool hit = RaytraceVoxel(viewRayPosition, viewHitIndex, viewRayDirection, viewRayPosition, viewHitIndex, dist, data);

            Material hitMaterial = hitMaterial(data, viewRayPosition, wavelength);

            return SpectrumToXYZ((hitMaterial.albedo * data.ambientOcclusion) / wavelengthPDF, wavelength) * xyzToRGBMatrix;
        #elif DISPLAY == 3
            for(int i = 0; i < 2; ++i) {
                VoxelData data;
                vec2 voxelmapCoord;
                float dist;
                ivec3 hitIndex = ivec3(floor(rayPosition));
                bool hit = RaytraceVoxel(rayPosition, hitIndex, rayDirection, rayPosition, hitIndex, dist, data);


                if(RandNextF() > throughput) {
                    throughput = 0.0;
                    break;
                }
                throughput /= saturate(throughput);

                if(!hit) {
                    estimate += throughput * 1.0;

                    break;
                } else {
                    rayPosition = rayPosition + data.flatNormal * 1e-5;
                    rayDirection = GenerateCosineVector(data.normalMap, RandNext2F());

                    throughput *= saturate(dot(rayDirection, data.normalMap)) * data.ambientOcclusion;
                    if(dot(rayDirection, data.normalMap) < 0.0) {
                        break;
                    }
                }
            }

            return vec3(estimate);
        #endif
    }
#endif