#if !defined GOOCH_SHADING
#define GOOCH_SHADING
    float GoochShading(in vec3 L, in vec3 V, in vec3 N, in float wavelength, in bool specular, in bool laterBounce) {
        vec3 cHighlight = vec3(20.0);
        vec3 cCool = BlackbodyRGB(70e3);
        vec3 cWarm = BlackbodyRGB(3000.0);

        float sHighlight;
        float sCool;
        float sWarm;
        RGBToSpectrum(sHighlight, wavelength, cHighlight.r, cHighlight.g, cHighlight.b, 0);
        RGBToSpectrum(sCool, wavelength, cCool.r, cCool.g, cCool.b, 0);
        RGBToSpectrum(sWarm, wavelength, cWarm.r, cWarm.g, cWarm.b, 0);

        float T = (dot(N, L) + 1.0) * 0.5;
        vec3 R = reflect(-L, N);
        float S = clamp((100.0 * dot(R, V) - 99.0), 0.0, 1.0);

        float gooch = mix(saturate(mix(sWarm, sCool, T * float(laterBounce))), sHighlight, S * float(specular));

        if(isnan(gooch)) {
            gooch = 0.0;
        }
        if(isinf(gooch)) {
            gooch = 1.0;
        }

        return gooch;
    }

    /*
                        //Hit the terrain
                        if(data.sameMaterial && data.id != 156) {
                            continue;
                        }

                        Material hitMaterial = hitMaterial(data, SceneSpaceFromVoxelSpace(rayPosition), wavelength);

                        mat3 tbnMatrix2 = GetRotationMatrix(data.tbnMatrix[2], data.flatNormal) * data.tbnMatrix;
                        vec3 viewDirection   = oldRayDirection * tbnMatrix2;
                            viewDirection.z = sign(dot(oldRayDirection, data.tbnMatrix[2])) * abs(viewDirection.z);

                        if(dot(data.flatNormal, celestialRayDirection) > 0.0 && !data.sameMaterial) {
                            VoxelData voxelDataLight;
                            vec3 hitPositionLight = hitMaterial.doRefract ? rayPosition : rayPosition - data.flatNormal * 1e-5;
                            ivec3 hitIndexLight = ivec3(floor(hitPositionLight));
                            vec2 voxelmapCoordLight;
                            float hitDistanceSurfaceLight;
                            bool hitLight = RaytraceVoxel(hitPositionLight, hitIndexLight, celestialRayDirection, hitPositionLight, hitIndexLight, hitDistanceSurfaceLight, voxelDataLight);

                            if(!hitLight) {
                                float bsdf = GoochShading(celestialRayDirection, -rayDirection, data.normalMap, wavelength, true, true) / pi;

                                float transmittance = EstimateTransmittance(rayPosition, celestialRayDirection, baseAttenuationCoefficients);

                                estimate += throughput * bsdf * irradiance * transmittance;
                            }
                        }

                        {
                            rayPosition = hitMaterial.emissive ? rayPosition : rayPosition + data.flatNormal * 1e-5;
                            hitIndex = ivec3(floor(rayPosition));

                            vec3 newDirection = GenerateUnitVector(RandNext2F());

                            throughput *= GoochShading(newDirection, -rayDirection, data.normalMap, wavelength, false, pathSegment < 4);

                            estimate += throughput * hitMaterial.emission;

                            if(isnan(throughput)) {
                                throughput = 0.0;
                            }
                            if(isinf(throughput)) {
                                throughput = 1.0;
                            }

                            rayDirection = newDirection;

                            if (pathSegment <= MAX_PATH_SEGMENTS) {
                                if (dot(rayDirection, data.tbnMatrix[2]) <= 0.0 && data.sameMaterial) { break; }
                            }
                        }
    */
#endif