float square(in float x) {
    return _square(x);
}
int square(in int x) {
    return _square(x);
}
vec2 square(in vec2 x) {
    return _square(x);
}
vec3 square(in vec3 x) {
    return _square(x);
}
vec4 square(in vec4 x) {
    return _square(x);
}

float cube(in float x) {
    return _cube(x);
}
int cube(in int x) {
    return _cube(x);
}
vec2 cube(in vec2 x) {
    return _cube(x);
}
vec3 cube(in vec3 x) {
    return _cube(x);
}
vec4 cube(in vec4 x) {
    return _cube(x);
}

float pow5(in float x) {
    return _pow5(x);
}
int pow5(in int x) {
    return _pow5(x);
}
vec2 pow5(in vec2 x) {
    return _pow5(x);
}
vec3 pow5(in vec3 x) {
    return _pow5(x);
}
vec4 pow5(in vec4 x) {
    return _pow5(x);
}

float pow6(in float x) {
    return _pow6(x);
}
int pow6(in int x) {
    return _pow6(x);
}
vec2 pow6(in vec2 x) {
    return _pow6(x);
}
vec3 pow6(in vec3 x) {
    return _pow6(x);
}
vec4 pow6(in vec4 x) {
    return _pow6(x);
}

float saturate(in float x) {
    return _saturate(x);
}
int saturate(in int x) {
    return _saturateInt(x);
}
vec2 saturate(in vec2 x) {
    return _saturate(x);
}
vec3 saturate(in vec3 x) {
    return _saturate(x);
}
vec4 saturate(in vec4 x) {
    return _saturate(x);
}

float minof(vec2 x) { 
    return min(x.x, x.y); 
}
float minof(vec3 x) { 
    return min(min(x.x, x.y), x.z); 
}
float minof(vec4 x) { 
    x.xy = min(x.xy, x.zw); return min(x.x, x.y); 
}

float maxof(vec2 x) { 
    return max(x.x, x.y); 
}
float maxof(vec3 x) { 
    return max(max(x.x, x.y), x.z); 
}
float maxof(vec4 x) { 
    x.xy = max(x.xy, x.zw); return max(x.x, x.y); 
}

float rcp(in int x) {
    float y = float(x);
    return _rcp(y);
}
float rcp(in float x) {
    return _rcp(x);
}
vec2 rcp(in vec2 x) {
    return _rcp(x);
}
vec3 rcp(in vec3 x) {
    return _rcp(x);
}
vec4 rcp(in vec4 x) {
    return _rcp(x);
}

float log10(in float x) {
    return _log10(x, 10.0);
}

int log10(in int x) {
    return int(_log10(x, 10.0));
}

vec2 log10(in vec2 x) {
    return _log10(x, 10.0);
}

vec3 log10(in vec3 x) {
    return _log10(x, 10.0);
}

vec4 log10(in vec4 x) {
    return _log10(x, 10.0);
}

float linearstep(in float x, float low, float high) {
    float data = x;
    float mapped = (data-low)/(high-low);

    return saturate(mapped);
}

vec2 linearstep(in vec2 x, float low, float high) {
    vec2 data = x;
    vec2 mapped = (data-low)/(high-low);

    return saturate(mapped);
}

vec3 linearstep(in vec3 x, float low, float high) {
    vec3 data = x;
    vec3 mapped = (data-low)/(high-low);

    return saturate(mapped);
}

vec4 linearstep(in vec4 x, float low, float high) {
    vec4 data = x;
    vec4 mapped = (data-low)/(high-low);

    return saturate(mapped);
}

vec2 sincos(float x) { return vec2(sin(x), cos(x)); }

#include "rotation.glsl"

vec2 CircleMap(in float index, in float count) {
    float goldenAngle = tau / ((sqrt(5.0) * 0.5 + 0.5) + 1.0);
    return vec2(cos(index * goldenAngle), sin(index * goldenAngle)) * sqrt(index / count);
}

vec2 CircleMap(in float point) {
    return vec2(cos(point), sin(point));
}

vec2 SpiralPoint(float angle, float scale) {
	return vec2(sin(angle), cos(angle)) * pow(angle / scale, 1.0 / (sqrt(5.0) * 0.5 + 0.5));
}

vec2 ConcentricSampleDisk(in vec2 rng) {
    vec2 uOffset = 2.0 * rng - 1.0;
    if(all(equal(uOffset, vec2(0.0)))) {
        return vec2(0.0);
    }
    float theta, r;
    if(abs(uOffset.x) > abs(uOffset.y)) {
        r = uOffset.x;
        theta = (pi / 4.0) * (uOffset.y / uOffset.x);
    } else {
        r = uOffset.x;
        theta = (pi / 2.0) - (pi / 4.0) * (uOffset.y / uOffset.x);
    }

    return r * vec2(cos(theta), sin(theta));
}

void ConstructCoordinateSystem(in vec3 v1, inout vec3 v2, inout vec3 v3) {
    if(abs(v1.x) > abs(v1.y)) {
        v2 = vec3(-v1.z, 0.0, v1.x) / length(v1);
    } else {
        v2 = vec3(0.0, v1.z, -v1.y) / length(v1);
    }
    v3 = cross(v1, v2);
}

vec3 GenerateUnitVector(vec2 hash) {
    hash.x *= tau; hash.y = hash.y * 2.0 - 1.0;
    return vec3(vec2(sin(hash.x), cos(hash.x)) * sqrt(saturate(1.0 - hash.y * hash.y)), hash.y);
}

vec3 GenerateCosineVector(vec3 vector, vec2 xy) {
    vec3 dir = GenerateUnitVector(xy);

    if(vector + dir == vec3(0.0)) {
        return vector;
    } else {
        return fNormalize(vector + dir);
    }
}

vec3 SampleRectangle(in vec3 Tx, in vec3 Ty, in vec2 radii, in float rng) {
    return radii.x * (2.0 * rng - 1.0) * Tx + radii.y * (2.0 * rng - 1.0) * Ty;
}

vec3 GenerateConeVector(vec3 vector, vec2 xy, float angle) {
    xy.x *= radians(360.0);
    float cosAngle = cos(angle);
    xy.y = xy.y * (1.0 - cosAngle) + cosAngle;
    vec3 sphereCap = vec3(vec2(cos(xy.x), sin(xy.x)) * sqrt(1.0 - xy.y * xy.y), xy.y);
    return Rotate(sphereCap, vec3(0, 0, 1), vector);
}

vec2 NormalDistributionRand(in float mean, in float sigma, in vec2 uv) {
    float r     = sqrt(-2.0 * log(uv.x)) * sigma;
    float theta = tau * uv.y;

    return (r * vec2(cos(theta), sin(theta))) + mean;
}

float NormalDistributionPDF(in float x, in float mean, in float standardDeviation) {
    return (1.0 / (standardDeviation * sqrt(2.0 * pi))) * exp(-0.5 * square((x - mean) / standardDeviation));
}

float NormalDistributionPDF(in float x, in float sigma) {
    return (1.0 / (sigma * sqrt(2.0 * pi))) * exp(-0.5 * square((x) / sigma));
}

vec2 SampleAsymmetricNormalDistribution(vec2 mean, vec2 standardDeviation1, vec2 standardDeviation2, vec2 xy) {
    float r = sqrt(-2.0 * log(1.0 - xy.x));
    float theta = radians(360.0) * xy.y;
    vec2 tmp = r * vec2(cos(theta), sin(theta));
    return mix(standardDeviation2, standardDeviation1, lessThan(tmp, vec2(0.0))) * tmp + mean;
}
float EstimateAsymmetricNormalDistribution(float x, float mean, float standardDeviation1, float standardDeviation2) {
    float norm = 2.0 / (standardDeviation1 * standardDeviation2 * (1.0 / standardDeviation1 + 1.0 / standardDeviation2) * sqrt(2.0 * pi));
    float tmp = (x - mean) / (x < mean ? standardDeviation1 : standardDeviation2);
    return norm * exp(-0.5 * tmp * tmp);
}

bool IntersectPlate(vec3 p, vec3 d, vec2 radii, out float t) {
    t = -p.z / d.z;

    const float epsilon = 2e-5;
    return -p.z*sign(d.z) > epsilon && all(lessThan(abs(p.xy + d.xy * t), radii));
}

vec2 RSI(vec3 pos, vec3 dir, float radius) {
    float radiusSquared = square(radius);
    float posDotDir = dot(pos, dir);
    float endDist = posDotDir*posDotDir + radiusSquared - dot(pos, pos);

    if(endDist < 0.0) return vec2(-1.0);

    endDist = sqrt(endDist);
    vec2 ret = -posDotDir + vec2(-endDist, endDist);

    return ret;
}

float RDI(vec3 rayPos, vec3 rayDir, in float radiusInner, in float radiusOuter) {
    if (rayPos.z == 0.0) { return 0.0; }
    if (sign(rayPos.z) == sign(rayDir.z)) { return -1.0; }
    
    float tPlane = -rayPos.z / rayDir.z;
    
    if (tPlane < 0.0) { return -1.0; }
    
    vec3 iPos = rayPos + rayDir * tPlane;
    float radius = length(iPos);
    
    if (radiusOuter < radius) { return -1.0; }
    if (radius < radiusInner) { return -1.0; }
    
    return tPlane;
}

bool IntersectSphere(in vec3 rayPosition, in vec3 rayDirection, in float radius, out float dist) {
    vec2 sphereDists = RSI(rayPosition, rayDirection, radius);
    
    if(sphereDists.x > 0.001) {
        dist = sphereDists.x;
        return true;
    } else {
        dist = sphereDists.y;
        if(dist < 0.001) {
            return false;
        } else {
            return true;
        }
    }
}

int LinePlaneIntersection(
    vec3 Lo, vec3 Lv,
    vec3 Po, vec3 Pn,
    out float t
) {
    float NdotV = dot(Pn, Lv);
    float NdotO = dot(Pn, Lo - Po);
    if (NdotV == 0.0) {
        if (NdotO == 0.0) {
            return 2;
        } else {
            return 0;
        }
    }

    t = NdotO / -NdotV;

    return 1;
}

int RayPlaneIntersection(
    vec3 Ro, vec3 Rv,
    vec3 Po, vec3 Pn,
    out float t
) {
    int tmp = LinePlaneIntersection(Ro, Rv, Po, Pn, t);
    if (tmp == 1 && t < 0.0) {
        return 0;
    }
    
    if (dot(Rv, Pn) > 0.0) { return 0; }

    return tmp;
}

int LineSphereIntersect(
	vec3 Lo, vec3 Lv,
	vec3 So, float Sr,
	out float t0, out float t1
) {
	float Rsquared = dot(Lo - So, Lo - So);
	float RMu      = dot(Lv, Lo - So) / length(Lv);

	float discriminant = RMu * RMu - Rsquared + Sr * Sr;

	if (discriminant < 0.0) {
		return 0;
	}

	if (discriminant == 0.0) {
		t0 = -RMu;
		return 1;
	}

	t0 = (-RMu - sqrt(discriminant)) / length(Lv);
	t1 = (-RMu + sqrt(discriminant)) / length(Lv);

	return 2;
}

vec4 CatRom(float x) {
	vec4 vec = vec4(1.0, x, x * x, x * x * x);
	const mat4 matrix = mat4(
		 0, 2, 0, 0,
		-1, 0, 1, 0,
		 2,-5, 4,-1,
		-1, 3,-3, 1
	);
	return (1.0/2.0) * matrix * vec;
}

vec4 BSpline(float x) {
    vec4 vec = vec4(1.0, x, x * x, x * x * x);
    const mat4 matrix = mat4(
            1, 4, 1, 0,
        -3, 0, 3, 0,
            3,-6, 3, 0,
        -1, 3,-3, 1
    );
    return (1.0/6.0) * matrix * vec;
}

vec4 textureBicubic(sampler2D sampler, vec2 coord) {
	vec2 res = textureSize(sampler, 0);

	coord = coord * res - 0.5;

	vec2 f = fract(coord);
	coord -= f;

	vec2 ff = f * f;
	vec4 w0;
	vec4 w1;
	w0.xz = 1 - f; w0.xz *= w0.xz * w0.xz;
	w1.yw = ff * f;
	w1.xz = 3 * w1.yw + 4 - 6 * ff;
	w0.yw = 6 - w1.xz - w1.yw - w0.xz;

	vec4 s = w0 + w1;
	vec4 c = coord.xxyy + vec2(-0.5, 1.5).xyxy + w1 / s;
	c /= res.xxyy;

	vec2 m = s.xz / (s.xz + s.yw);
	return mix(
		mix(textureLod(sampler, c.yw, 0), textureLod(sampler, c.xw, 0), m.x),
		mix(textureLod(sampler, c.yz, 0), textureLod(sampler, c.xz, 0), m.x),
		m.y);
}

// The following code is licensed under the MIT license: https://gist.github.com/TheRealMJP/bc503b0b87b643d3505d41eab8b332ae
vec4 textureCatmullRom(in sampler2D tex, vec2 uv) {
    vec2 texSize = textureSize(tex, 0);
    // We're going to sample a a 4x4 grid of texels surrounding the target UV coordinate. We'll do this by rounding
    // down the sample location to get the exact center of our "starting" texel. The starting texel will be at
    // location [1, 1] in the grid, where [0, 0] is the top left corner.
    vec2 samplePos = uv * texSize;
    vec2 texPos1 = floor(samplePos - 0.5) + 0.5;

    // Compute the fractional offset from our starting texel to our original sample location, which we'll
    // feed into the Catmull-Rom spline function to get our filter weights.
    vec2 f = samplePos - texPos1;

    // Compute the Catmull-Rom weights using the fractional offset that we calculated earlier.
    // These equations are pre-expanded based on our knowledge of where the texels will be located,
    // which lets us avoid having to evaluate a piece-wise function.
    vec2 w0 = f * ( -0.5 + f * (1.0 - 0.5*f));
    vec2 w1 = 1.0 + f * f * (-2.5 + 1.5*f);
    vec2 w2 = f * ( 0.5 + f * (2.0 - 1.5*f) );
    vec2 w3 = f * f * (-0.5 + 0.5 * f);
    
    // Work out weighting factors and sampling offsets that will let us use bilinear filtering to
    // simultaneously evaluate the middle 2 samples from the 4x4 grid.
    vec2 w12 = w1 + w2;
    vec2 offset12 = w2 / (w1 + w2);

    // Compute the final UV coordinates we'll use for sampling the texture
    vec2 texPos0 = texPos1 - vec2(1.0);
    vec2 texPos3 = texPos1 + vec2(2.0);
    vec2 texPos12 = texPos1 + offset12;

    texPos0 /= texSize;
    texPos3 /= texSize;
    texPos12 /= texSize;

    vec4 result = vec4(0.0);
    result += texture(tex, vec2(texPos0.x,  texPos0.y)) * w0.x * w0.y;
    result += texture(tex, vec2(texPos12.x, texPos0.y)) * w12.x * w0.y;
    result += texture(tex, vec2(texPos3.x,  texPos0.y)) * w3.x * w0.y;

    result += texture(tex, vec2(texPos0.x,  texPos12.y)) * w0.x * w12.y;
    result += texture(tex, vec2(texPos12.x, texPos12.y)) * w12.x * w12.y;
    result += texture(tex, vec2(texPos3.x,  texPos12.y)) * w3.x * w12.y;

    result += texture(tex, vec2(texPos0.x,  texPos3.y)) * w0.x * w3.y;
    result += texture(tex, vec2(texPos12.x, texPos3.y)) * w12.x * w3.y;
    result += texture(tex, vec2(texPos3.x,  texPos3.y)) * w3.x * w3.y;

    return result;
}
//End of MIT code

float SignExtract(float x) {
	return uintBitsToFloat((floatBitsToUint(x) & 0x80000000u) | floatBitsToUint(1.0));
}
vec3 SignExtract(vec3 x) {
	return uintBitsToFloat((floatBitsToUint(x) & 0x80000000u) | floatBitsToUint(1.0));
}

#define NEWTON_ITER 8
#define HALLEY_ITER 8

float cbrt(float x) {
	float y = sign(x) * uintBitsToFloat(floatBitsToUint(abs(x)) / 3u + 0x2a514067u);

	for(int i = 0; i < NEWTON_ITER; ++i)
    	y = (2. * y + x / (y * y)) * .333333333;

    for(int i = 0; i < HALLEY_ITER; ++i) {
    	float y3 = y * y * y;
        y *= (y3 + 2. * x) / (2. * y3 + x);
    }
    
    return y;
}

float erf(float x) {
    // constants
    float a1 =  0.254829592;
    float a2 = -0.284496736;
    float a3 =  1.421413741;
    float a4 = -1.453152027;
    float a5 =  1.061405429;
    float p  =  0.3275911;
 
    // Save the sign of x
    int sign = 1;
    if (x < 0)
        sign = -1;
    x = abs(x);
 
    // A&S formula 7.1.26
    float t = 1.0/(1.0 + p*x);
    float y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*exp(-x*x);
 
    return sign*y;
}

float erfinv(float x) {
    float w, p;
    w = - log((1.0f-x)*(1.0f+x));
    if (w < 5.000000) {
        w = w - 2.500000;
        p = 2.81022636e-08;
        p = 3.43273939e-07 + p*w;
        p = -3.5233877e-06 + p*w;
        p = -4.39150654e-06 + p*w;
        p = 0.00021858087 + p*w;
        p = -0.00125372503 + p*w;
        p = -0.00417768164 + p*w;
        p = 0.246640727 + p*w;
        p = 1.50140941 + p*w;
    } else {
        w = sqrt(w) - 3.000000;
        p = -0.000200214257;
        p = 0.000100950558 + p*w;
        p = 0.00134934322 + p*w;
        p = -0.00367342844 + p*w;
        p = 0.00573950773 + p*w;
        p = -0.0076224613 + p*w;
        p = 0.00943887047 + p*w;
        p = 1.00167406 + p*w;
        p = 2.83297682 + p*w;
    }
    return p*x;
}

vec3 refractIT(vec3 I, vec3 N, const float iori, const float iort) {
    float eta = iori / iort;
    if(dot(N, I) > 0.0) {
        eta = iort / iori;
    }
    float k = 1.0 - eta * eta * (1.0 - dot(N, I) * dot(N, I));
    if(k < 0.0) {
        return vec3(0.0);
    } else {
        return eta * I - (eta * dot(N, I) + sqrt(k)) * N;
    }
}

float Plancks(float t, float lambda) {
    const float h = 6.62607015e-16;
    const float c = 2.9e17;
    const float k = 1.38e-5;

    float p1 = 2.0 * h * pow(c, 2.0) * pow(lambda, -5.0);
    float p2 = exp((h * c) / (lambda * k * t)) - 1.0;

    return p1 / p2;
}

vec3 BlackbodyRGB(float temperature) {
    vec3 rgb = vec3(Plancks(temperature, 680.0), Plancks(temperature, 555.0), Plancks(temperature, 425.0));
            rgb = rgb / max(rgb.x, max(rgb.y, rgb.z));

    return rgb;
}

uniform sampler2D colortex11;

float scale;

float xFit_1931(float wave) {
    float t1 = (wave - 442.0f)*((wave < 442.0f) ? 0.0624f : 0.0374f);
    float t2 = (wave - 599.8f)*((wave < 599.8f) ? 0.0264f : 0.0323f);
    float t3 = (wave - 501.1f)*((wave < 501.1f) ? 0.0490f : 0.0382f);
    return 0.362f*exp(-0.5f*t1*t1) + 1.056f*exp(-0.5f*t2*t2) - 0.065f*exp(-0.5f*t3*t3);
}
float yFit_1931(float wave) {
    float t1 = (wave - 568.8f)*((wave < 568.8f) ? 0.0213f : 0.0247f);
    float t2 = (wave - 530.9f)*((wave < 530.9f) ? 0.0613f : 0.0322f);
    return 0.821f*exp(-0.5f*t1*t1) + 0.286f*exp(-0.5f*t2*t2);
}
float zFit_1931(float wave) {
    float t1 = (wave - 437.0f)*((wave < 437.0f) ? 0.0845f : 0.0278f);
    float t2 = (wave - 459.0f)*((wave < 459.0f) ? 0.0385f : 0.0725f);

    return 1.217f*exp(-0.5f*t1*t1) + 0.681f*exp(-0.5f*t2*t2);
}

vec3 SpectrumToXYZ_1931(in float spectrum, in float w) {
    float n = (w - 390.0);
    int i = int(n);

    int n0 = int(n);
    int n1 = min(n0 + 1, 440);

    vec3 xyz = vec3(xFit_1931(w), yFit_1931(w), zFit_1931(w));

    xyz = vec3(spectrum * scale) * xyz;

    return xyz * float(441.0) / 113.042;
}

vec3 Tech_ComputeApproximateCIE_XYZ(in float wavelength) {
    float n = wavelength - 390.0;
    vec3 xyz;

    float bCoord    = n - 81.0;
    float gCoord    = (n - 166.0) * 0.016;
    gCoord          = square(gCoord) * sqrt(abs(gCoord)) * (gCoord < 0.0 ? 1.72 : 0.8);

    xyz.r  = exp(-square((n - 209.0) * 0.022)) * 1.151326;
    xyz.r += exp(-square((n - 156.0) * 0.050)) * 5.75663E-02;

    xyz.g += exp(-gCoord) * 9.999930E-01;
    xyz.g += exp(-square(n - 86.0) * 0.0007) * 9.999930E-02;

    xyz.b += exp(-square((n - 54.5) * 0.035)) * 1.934819;
    xyz.b += exp(-square(bCoord * (bCoord < 0.0 ? 0.105 : 0.0292))) * 4.4500837E-01;
    xyz.b *= 1.0 - exp(-square(n) * 0.005);
    xyz.b  = bCoord < 0.0 ? xyz.b * 0.982 + 0.018 : xyz.b;

    xyz.r += exp(-square((n - 51.0) * 0.04)) * 3.494153E-01;
    xyz.r += exp(-square((n - 79.0) * 0.07)) * 1.0482459E-01;
    xyz.r  = bCoord < 0.0 ? min(xyz.r, xyz.b * 0.19) : xyz.r;

    return xyz;
}

vec3 SpectrumToXYZFit_CIE2012_Tech(in float spectrum, in float w) {
    vec3 xyz = Tech_ComputeApproximateCIE_XYZ(w);

    return xyz * spectrum * scale * float(441.0) / 113.042;
}

vec3 SpectrumToXYZ(in float spectrum, in float w) {
    float n = (w - 390.0);
    int i = int(n);

    int n0 = int(n);
    int n1 = min(n0 + 1, 440);

    vec3 cie0 = texelFetch(colortex11, ivec2(n0, 1), 0).xyz;
    vec3 cie1 = texelFetch(colortex11, ivec2(n1, 1), 0).xyz;

    vec3 xyz = mix(cie0, cie1, fract(n));

    xyz = vec3(spectrum * scale) * xyz;

    return xyz * float(441.0) / 113.042;
}

vec3 ComputeApproximateCIE_XYZ(in float wavelength) {
    float x = EstimateAsymmetricNormalDistribution(wavelength, peakWavelengths.x, xStandardDeviations.x, xStandardDeviations.y) * xyzStrengths.x;
    float y = EstimateAsymmetricNormalDistribution(wavelength, peakWavelengths.y, yStandardDeviations.x, yStandardDeviations.y) * xyzStrengths.y;
    float z = EstimateAsymmetricNormalDistribution(wavelength, peakWavelengths.z, zStandardDeviations.x, zStandardDeviations.y) * xyzStrengths.z;

    return vec3(x, y, z);
}

vec3 SpectrumToXYZ(in vec3 spectrum, in float w) {
    float n = (w - 390.0);
    int i = int(n);
    if(i < 0 || i >= (830 - 390)) {
        return vec3(0.0);
    } else {
        vec3 xyz = ComputeApproximateCIE_XYZ(w);
        
        xyz = xyz * spectrum * scale;

        return xyz * float(441.0) / vec3(113.042);
    }
}

float Air(in float wl) {
    const float kelvin = 273.15; //Conversion to get temperature in kelvin from celcius.
    const float pressure = 1013.2; //Millibars
    const float temperature = 13.9; //Celcius
    const float T = temperature + kelvin; //Kelvin
    float vaporPressure = 0.61094 * exp(17.625 * temperature / (temperature + 243.04)); //Kilopascal
    vaporPressure = vaporPressure / 10.0; //Convert to millibars
    return 1.0 + (77.6e-6 / T) * (1.0 + (7.52e-3 / pow(wl, 2.0))) * (pressure + 4810.0 * (vaporPressure / T));
}

float Plastic(in float wl) {
    return 1.5151+0.00279*pow(wl,-2.0)+5.0756E-4*pow(wl,-4.0);
}

float Glass(in float wl) {
    return sqrt(1.0+1.03961212/(1.0-0.00600069867/pow(wl,2.0))+0.231792344/(1.0-0.0200179144/pow(wl,2.0))+1.01046945/(1.0-103.560653/pow(wl,2.0)));
}

float SellmeierEquation(in float wavelength) {
    wavelength *= 1e-3;
    float wavelength2 = square(wavelength);
    const float B1 = 1.03961212;
    const float B2 = 0.23179234;
    const float B3 = 1.01046945;
    const float C1 = 6.00069e-3;
    const float C2 = 2.00179e-2;
    const float C3 = 1.035606e2;
    return 1.0 + ((B1 * wavelength2) / (wavelength2 - C1)) + ((B2 * wavelength2) / (wavelength2 - C2)) + ((B3 * wavelength2) / (wavelength2 - C3));
}

uniform sampler2D colortex12;

float InterpolateSpectrum(float wavelength, int lutRow) {
	if (wavelength < 380.000000 || wavelength > 720.000000) {
		return 0.0;
	}

	float fractionalIndex = float(textureSize(colortex12, 0).x - 1) * (wavelength - 380.0) / (720.0 - 380.0);
	int indexBelow = int(floor(fractionalIndex));
	int indexAbove = int(ceil (fractionalIndex));
	return mix(
		texelFetch(colortex12, ivec2(indexBelow, lutRow), 0).x,
		texelFetch(colortex12, ivec2(indexAbove, lutRow), 0).x,
		fract(fractionalIndex)
	);
}

//Mitsuba Based Code
//This code is based directly on code from Mitsuba, the only real differences are how I implemented them and porting them to GLSL.
void RGBToSpectrum(out float result, in float wl, in float r, in float g, in float b, in int intent) {
    result = 0.0;

    if(intent == 0) {
        float rgbToSpecWhite = InterpolateSpectrum(wl, 0);
        float rgbToSpecCyan = InterpolateSpectrum(wl, 1);
        float rgbToSpecMagenta = InterpolateSpectrum(wl, 2);
        float rgbToSpecYellow = InterpolateSpectrum(wl, 3);
        float rgbToSpecRed = InterpolateSpectrum(wl, 4);
        float rgbToSpecGreen = InterpolateSpectrum(wl, 5);
        float rgbToSpecBlue = InterpolateSpectrum(wl, 6);
        if (r <= g && r <= b) {
            result += r * rgbToSpecWhite;

            if (g <= b) {
                result += (g - r) * rgbToSpecCyan;
                result += (b - g) * rgbToSpecBlue;
            } else {
                result += (b - r) * rgbToSpecCyan;
                result += (g - b) * rgbToSpecGreen;
            }
        } else if (g <= r && g <= b) {
            result += g * rgbToSpecWhite;

            if (r <= b) {
                result += (r - g) * rgbToSpecMagenta;
                result += (b - r) * rgbToSpecBlue;
            } else {
                result += (b - g) * rgbToSpecMagenta;
                result += (r - b) * rgbToSpecRed;
            }
        } else {
            result += b * rgbToSpecWhite;

            if (r <= g) {
                result += (r - b) * rgbToSpecYellow;
                result += (g - r) * rgbToSpecGreen;
            } else {
                result += (g - b) * rgbToSpecYellow;
                result += (r - g) * rgbToSpecRed;
            }
        }

        result *= .94;
    } else {
        float rgbToIllumWhite = InterpolateSpectrum(wl, 7);
        float rgbToIllumCyan = InterpolateSpectrum(wl, 8);
        float rgbToIllumMagenta = InterpolateSpectrum(wl, 9);
        float rgbToIllumYellow = InterpolateSpectrum(wl, 10);
        float rgbToIllumRed = InterpolateSpectrum(wl, 11);
        float rgbToIllumGreen = InterpolateSpectrum(wl, 12);
        float rgbToIllumBlue = InterpolateSpectrum(wl, 13);
        if (r <= g && r <= b) {
            result += r * rgbToIllumWhite;

            if (g <= b) {
                result += (g - r) * rgbToIllumCyan;
                result += (b - g) * rgbToIllumBlue;
            } else {
                result += (b - r) * rgbToIllumCyan;
                result += (g - b) * rgbToIllumGreen;
            }
        } else if (g <= r && g <= b) {
            result += g * rgbToIllumWhite;

            if (r <= b) {
                result += (r - g) * rgbToIllumMagenta;
                result += (b - r) * rgbToIllumBlue;
            } else {
                result += (b - g) * rgbToIllumMagenta;
                result += (r - b) * rgbToIllumRed;
            }
        } else {
            result += b * rgbToIllumWhite;

            if (r <= g) {
                result += (r - b) * rgbToIllumYellow;
                result += (g - r) * rgbToIllumGreen;
            } else {
                result += (g - b) * rgbToIllumYellow;
                result += (r - g) * rgbToIllumRed;
            }
        }

        result *= .86445;
    }
    
    if(isnan(result)) {
        result = 1.0;
    }
}
//End of Mistuba code