#if !defined LIB_SUN_MOON_CALCULATION
#define LIB_SUN_MOON_CALCULATION
    const float dayMs = 8.64e+7;
    const float J1970 = 2440588.0;
    const float J2000 = 2451545.0;
    const float rad = 0.0174532925;
    const float ob_e = rad * 23.4397;

    float ToJulian(in float date) {
        return date / dayMs - 0.5 + J1970;
    }
    float FromJulian(in float J) {
        return 0.0;
    }
    float ToDays(in float date) {
        return ToJulian(date) - J2000;
    }

    float RightAscension(in float l, in float b) {
        return atan(sin(l) * cos(ob_e) - tan(b) * sin(ob_e), cos(l));
    }
    float Declination(in float l, in float b) {
        float val = asin(sin(b) * cos(ob_e) + cos(b) * sin(ob_e) * sin(l));
        return val;
    }

    float Azimuth(in float H, in float phiT, in float dec) {
        return atan(sin(H), cos(H) * sin(phiT) - tan(dec) * cos(phiT));
    }
    float Altitude(in float H, in float phiT, in float dec) {
        return asin(sin(phiT) * sin(dec) + cos(phiT) * cos(dec) * cos(H));
    }

    float SiderealTime(in float d, in float lw) {
        return rad * (280.16 + 360.9856235 * d) - lw;
    }

    float AstroRefraction(in float h) {
        if (h < 0.0) {
            h = 0.0;
        }

        return 0.0002967 / tan(h + 0.00312536 / (h + 0.08901179));
    }

    float SolarMeanAnomaly(in float d) {
        return rad * (357.5291 + 0.98560028 * d);
    }

    float EclipticLongitude(in float M) {
        float C = rad * (1.9148 * sin(M) + 0.02 * sin(2 * M) + 0.0003 * sin(3 * M)),
              P = rad * 102.9372;

        return M + C + P + pi;
    }

    vec2 SunCoords(in float d) {
        float M = SolarMeanAnomaly(d),
              L = EclipticLongitude(M);

        return vec2(
            Declination(L, 0.0),
            RightAscension(L, 0.0)
        );
    }
    vec3 MoonCoords(in float d) {
        float L = rad * (218.316 + 13.176396 * d),
              M = rad * (134.963 + 13.064993 * d),
              F = rad * (93.272 + 13.229350 * d);

        float l = L + (rad * 6.289) * sin(M),
              b = rad * 5.128 * sin(F),
              dt = 385001 - 20905 * cos(M);

        return vec3(
            Declination(l, b),
            RightAscension(l, b),
            dt
        );
    }

    float parallacticAngle;

    vec3 SunPosition(in float date, in float lat, in float lng) {
        float lw   = rad * -lng,
              phiT = rad * lat,
              d    = ToDays(date);

        vec2 c = SunCoords(d);
        float H = SiderealTime(d, lw) - c.y;

        return vec3(Azimuth(H, phiT, c.x), Altitude(H, phiT, c.x), 0.0);
    }
    vec3 MoonPosition(in float date, in float lat, in float lng) {
        float lw   = rad * -lng,
              phiT = rad * lat,
              d    = ToDays(date);

        vec3 c = MoonCoords(d);
        float H = SiderealTime(d, lw) - c.y;
        float h = Altitude(H, phiT, c.x);
              //h = h + AstroRefraction(h);

        parallacticAngle = atan(sin(H), tan(phiT) * cos(c.x) - sin(c.x) * cos(H));

        return vec3(Azimuth(H, phiT, c.x), h, c.z);
    }
#endif