#if !defined LIB_VOXELIZATION
#define LIB_VOXELIZATION
    const uvec2 minBounds = uvec2(0, 0);
    const uvec2 maxBounds = uvec2(shadowMapQuality / 20, 384) - 1;

    vec3 VoxelSpaceFromSceneSpace(in vec3 position) {
        position.y += 64.0;
        const int radius = shadowMapQuality / 40;

        position.xz += fract(cameraPosition.xz) + radius;
        position.y  += cameraPosition.y;

        return position;
    }

    vec3 SceneSpaceFromVoxelSpace(in vec3 position) {
        position.y -= 64.0;
        const int radius = shadowMapQuality / 40;

        position.xz -= fract(cameraPosition.xz) + radius;
        position.y  -= cameraPosition.y;

        return position;
    }

    bool InsideVoxelmap(uvec3 idx) {
        const uvec2 maxBounds = uvec2(uint(shadowMapQuality) / 20u, 384u) - 1u;

        return all(lessThanEqual(idx, maxBounds.xyx));
    }

    bool InsideVoxelmap(ivec3 idx) {
        return InsideVoxelmap(uvec3(idx));
    }

    bool InsideVoxelmap(vec3 idx) {
        const vec2 maxBounds = vec2(shadowMapQuality / 20, 384) - 1.0;

        return all(greaterThanEqual(idx, vec3(0.0))) && all(lessThanEqual(idx, maxBounds.xyx));
    }

    bool OutsideVoxelmap(uvec3 idx) {
        const uvec2 maxBounds = uvec2(uint(shadowMapQuality) / 20u, 384u) - 1u;

        return any(greaterThan(idx, maxBounds.xyx));
    }

    bool OutsideVoxelmap(ivec3 idx) {
        return OutsideVoxelmap(uvec3(idx));
    }

    bool OutsideVoxelmap(vec3 idx) {
        const vec2 maxBounds = vec2(shadowMapQuality / 20, 384) - 1.0;

        return any(lessThan(idx, vec3(0.0))) && any(greaterThan(idx, maxBounds.xyx));
    }

    uvec2 CalculateVoxelmapIdx(uvec3 idx) {
        const uint radius = uint(shadowMapQuality) / 20u;

        uvec2 tileIdx = uvec2(idx.y % 20u, idx.y / 20u);
        uvec2 position = idx.xz + radius * tileIdx;

        return position;
    }

    ivec2 CalculateVoxelmapIdx(ivec3 idx) {
        const int radius = shadowMapQuality / 20;

        ivec2 tileIdx = ivec2(idx.y % 20, idx.y / 20);
        ivec2 position = idx.xz + radius * tileIdx;

        return position;
    }

    vec2 CalculateVoxelmapCoord(vec3 idx) {
        vec2 position = CalculateVoxelmapIdx(ivec3(idx));

        position /= shadowMapQuality;
        position += 0.5 / shadowMapQuality;

        return position;
    }
#endif