#version 450
#extension GL_ARB_gpu_shader_int64 : enable
#extension AMD_gpu_shader_int64 : enable

layout (r32ui) uniform uimage2D colorimg4;

layout (local_size_x = 16, local_size_y = 16) in;

uniform sampler2D colortex0;

uniform float viewWidth, viewHeight;

#include "/lib/utility/universal.glsl"

shared uint histogramShared[256];

float GetLuminance(vec3 color) {
    return dot(color, lumacoeff_rec709);
}

uint HDRToHistogramBin(vec3 hdrColor) {
    float luminance = GetLuminance(hdrColor);

    float logLuminance = saturate((log2(luminance) - -10.0) * (1.0 / 28.0));
    return uint(logLuminance * 254.0 + 1.0);
}

void main() {
    ivec2 fragPos = ivec2(gl_GlobalInvocationID.xy);

    histogramShared[gl_LocalInvocationIndex] = 0u;

    groupMemoryBarrier();
    barrier();

    if(int(gl_GlobalInvocationID.x) < viewWidth && int(gl_GlobalInvocationID.y) < viewHeight) {
        vec3 hdrColor = texelFetch(colortex0, fragPos, 0).rgb;
        uint binIndex = HDRToHistogramBin(hdrColor);
        atomicAdd(histogramShared[binIndex], 1);
    }

    memoryBarrierShared();
    barrier();

    imageAtomicAdd(colorimg4, ivec2(int(gl_LocalInvocationIndex), 0), histogramShared[gl_LocalInvocationIndex]);
}