#version 450
#extension GL_ARB_gpu_shader_int64 : enable
#extension AMD_gpu_shader_int64 : enable

layout (rgba32f) uniform image2D colorimg3;

layout (local_size_x = 16, local_size_y = 16) in;
const ivec3 workGroups = ivec3(1, 1, 1);

uniform usampler2D colortex4;

#include "/lib/utility/universal.glsl"

shared float histogramShared[256];
shared float weightedSumArr[256];

const float K = 27.0;
const float calibration = 2.0 * K / 100.0;

const float minExposure = tau / 120e3;
const float maxExposure = tau;

float ComputeEV100FromAverageLuminance(in float averageLuminance) {
    return log2(averageLuminance * calibration);
}

float EV100ToExposure(in float EV100) {
    float maxLuminance = calibration * pow(2.0, EV100);

    return 1.0 / maxLuminance;
}

void main() {
    float countForThisBin = float(texelFetch(colortex4, ivec2(int(gl_LocalInvocationIndex), 0), 0).r);
    histogramShared[gl_LocalInvocationIndex] = countForThisBin;

    groupMemoryBarrier();
    barrier();

    float cmf[256];
    cmf[0] = histogramShared[0];
    for(int n = 1; n < 256; ++n) {
        cmf[n] = cmf[n - 1] + histogramShared[n];
    }

    float maxCmf = cmf[255];

    for(int n = 0; n < 256; ++n) {
        cmf[n] = cmf[n] / maxCmf;
    }

    const float ignoreDark   = 0.7;
    const float ignoreBright = 1.0;

    float binWeight = clamp(cmf[gl_LocalInvocationIndex], ignoreDark, ignoreBright);
    if (gl_LocalInvocationIndex > 0) {
        binWeight -= clamp(cmf[gl_LocalInvocationIndex - 1], ignoreDark, ignoreBright);
    } else {
        binWeight -= ignoreDark;
    }

    weightedSumArr[gl_LocalInvocationIndex] = binWeight * float(gl_LocalInvocationIndex);

    groupMemoryBarrier();
    barrier();

    for(uint histogramSampleIndex = (256 >> 1); histogramSampleIndex > 0; histogramSampleIndex /= 2) {
        if(gl_LocalInvocationIndex < histogramSampleIndex) {
            weightedSumArr[gl_LocalInvocationIndex] += weightedSumArr[gl_LocalInvocationIndex + histogramSampleIndex];
        }

        groupMemoryBarrier();
        barrier();
    }

    if(gl_LocalInvocationIndex == 0) {
        const float bias = 0.0;
        float weightedLogAverage = (weightedSumArr[0] / (ignoreBright - ignoreDark)) - 1.0;
        float weightedAverageLuminance = exp2(((weightedLogAverage / 254.0) * 28.0) + -10.0);
        float exposure = EV100ToExposure(ComputeEV100FromAverageLuminance(weightedAverageLuminance) - bias);
              exposure = clamp(exposure, minExposure, maxExposure + bias);
        imageStore(colorimg3, ivec2(0, 0), vec4(exposure, 0.0, 0.0, 0.0));
    }
}