#version 450
#extension GL_ARB_gpu_shader_int64 : enable
#extension AMD_gpu_shader_int64 : enable

#define STAGE_FRAGMENT

layout(location = 0) out vec3 color;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform usampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D colortex9;
uniform sampler1D colortex10;
uniform sampler2D colortex14;

uniform sampler1D colortex15;

uniform sampler2D shadowtex0;

uniform int frameCounter;

uniform vec2 viewSize;

layout(location = 0) in vec2 textureCoordinate;

/*
    const int colortex0Format  = RGBA32F; //Main scene color
    const int colortex1Format  = RGBA32F;
    const int colortex2Format  = RGBA32F;
    const int colortex3Format  = RGBA32F; //Exposure
    const int colortex4Format  = R32UI; //Histogram
    const int colortex5Format  = RGBA32F; //Misc Purkinje Stuff
    const int colortex6Format  = RGBA16F; //Albedo
    const int colortex7Format  = RGBA16F; //Normal
    const int colortex8Format  = RGBA16F; //Specular
    const int colortex9Format  = RGB32F; //Specular

    const int shadowcolor1Format = R32UI;
*/

#include "/lib/utility/universal.glsl"
#include "/lib/fragment/aces/rrt.glsl"
#include "/lib/rng/bayer.glsl"

#include "/lib/fragment/tonemaps/misc.glsl"
#include "/lib/fragment/tonemaps/piecewiseFilmic.glsl"
#include "/lib/fragment/tonemaps/zombyes.glsl"

float Luminance(in vec3 v) {
    return dot(v, lumacoeff_rec709);
}

vec3 ChangeLuminance(in vec3 color, in float luminanceOut) {
    float luminanceIn = Luminance(color);
    return color * (luminanceOut / luminanceIn);
}

vec3 ReinhardJodie(in vec3 color) {
    vec3 tv = color / (1.0 + color);
    return mix(color / (1.0 + Luminance(color)), tv, tv);
}

vec3 ReinhardLuminance(in vec3 color) {
    float luminanceOld = Luminance(color);
    float numerator = luminanceOld * (1.0 + (luminanceOld / square(48.0)));
    float luminanceNew = numerator / (1.0 + luminanceOld);
    return ChangeLuminance(color, luminanceNew);
}

vec3 Reinhard(in vec3 color) {
    return color / (1.0 + color);
}

//https://github.com/TheRealMJP/BakingLab/blob/master/BakingLab/ACES.hlsl
const mat3x3 ACESInputMat = mat3(
    0.59719, 0.35458, 0.04823,
    0.07600, 0.90834, 0.01566,
    0.02840, 0.13383, 0.83777
);

const mat3x3 ACESOutputMat = mat3(
     1.60475, -0.53108, -0.07367,
    -0.10208,  1.10813, -0.00605,
    -0.00327, -0.07276,  1.07602
);

vec3 RRTAndODTFit(vec3 v) {
    vec3 a = v * (v + 0.0245786f) - 0.000090537f;
    vec3 b = v * (0.983729f * v + 0.4329510f) + 0.238081f;
    return a / b;
}

vec3 ACESFitted(vec3 color) {
    color = color * ACESInputMat;
    color = RRTAndODTFit(color);
    color = color * ACESOutputMat;
    color = saturate(color);

    return color;
}

void ApplyTonemap(inout vec3 color, in float exposure) {
    #if TONEMAP_OPERATOR == 0
        color = color * exposure;
        color = academy_color_encoding_sytem_rrt(color);
        color = LinearToSrgb(color);
    #elif TONEMAP_OPERATOR == 1
        color = color * exposure;
        color = Tonemap(color);
        color = LinearToSrgb(color);
    #elif TONEMAP_OPERATOR == 2
        color = color * exposure;
        color = apply_filmic_curve(color);
        color = LinearToSrgb(color);
    #elif TONEMAP_OPERATOR == 3
        color = color * exposure;
        color = ReinhardLuminance(color);
        color = LinearToSrgb(color);
    #elif TONEMAP_OPERATOR == 4
        color = color * exposure;
        color = TonemapHejlBurgess(color);
    #elif TONEMAP_OPERATOR == 5
        color = color * exposure;
        color = TonemapZombyeOld(color);
    #elif TONEMAP_OPERATOR == 6
        color = color * exposure;
        color = 1.0 - exp(-color);
        color = LinearToSrgb(color);
    #elif TONEMAP_OPERATOR == 7
        color = color * exposure;
        color = LinearToSrgb(color);
    #endif
}

void main() {
    vec2 coordinate = 1.0 - textureCoordinate;
    #ifdef LEGACY_LENS_SYSTEM
        coordinate = textureCoordinate;
    #endif

    #if DISPLAY == 0
        color = max(texture(colortex0, coordinate).rgb, 0.0);
        //color += abs(complexAbs(complexVec3(texture(colortex1, textureCoordinate).rgb, texture(colortex2, textureCoordinate).rgb)));

        float exposure = texelFetch(colortex3, ivec2(0, 0), 0).r * rcp(pi);
        const float m1 = -1.0;
        const float m2 = -6.0;

        const vec3 rodResponseCoefficients = vec3(1e-2, 0.2, 0.1);

        float rodResponse = dot(color, rodResponseCoefficients);
        float p = (log2(rodResponse) - m2) / (m1 - m2);
              p = 1.0 - saturate(p);

        vec3 tint = vec3(1.0);
        color = mix(color, rodResponse * tint, p);

        ApplyTonemap(color, exposure);
    #elif DISPLAY == 1
        color = texture(colortex0, coordinate).rgb * 0.5 + 0.5;
    #elif DISPLAY == 2
        color = texture(colortex0, coordinate).rgb * (1.0 / 120e3);
        color = LinearToSrgb(color);
    #elif DISPLAY == 3
        color = texture(colortex0, coordinate).rgb;
        color = LinearToSrgb(color);
    #endif

    color += Bayer32(gl_FragCoord.st) / 256.0;

    //color = texture(colortex14, textureCoordinate).rgb * 0.1;

    /*
    uint bin = uint(gl_FragCoord.x) / 2u;
    if (bin < 256) {
        float pdf_bin = float(texelFetch(colortex4, ivec2(bin, 0), 0).x);
        if (gl_FragCoord.y < ((256.0 / (viewSize.x * viewSize.y)) * pdf_bin)) {
            color = vec3(1.0);
        } else if (gl_FragCoord.y < 256.0) {
            color = vec3(0.0);
        }
    }
    //*/
}